import weakref
from functools import wraps

import numpy as np

from yt.data_objects.image_array import ImageArray
from yt.frontends.ytdata.utilities import save_as_dataset
from yt.funcs import get_output_filename, iter_fields, mylog
from yt.loaders import load_uniform_grid
from yt.utilities.lib.api import CICDeposit_2, add_points_to_greyscale_image
from yt.utilities.lib.pixelization_routines import pixelize_cylinder
from yt.utilities.on_demand_imports import _h5py as h5py

from .fixed_resolution_filters import apply_filter, filter_registry
from .volume_rendering.api import off_axis_projection


class FixedResolutionBuffer:
    r"""
    FixedResolutionBuffer(data_source, bounds, buff_size, antialias = True)

    This accepts a 2D data object, such as a Projection or Slice, and
    implements a protocol for generating a pixelized, fixed-resolution
    image buffer.

    yt stores 2D AMR data internally as a set of 2D coordinates and the
    half-width of individual pixels.  Converting this to an image buffer
    requires a deposition step, where individual variable-resolution pixels
    are deposited into a buffer of some resolution, to create an image.
    This object is an interface to that pixelization step: it can deposit
    multiple fields.  It acts as a standard YTDataContainer object, such that
    dict-style access returns an image of a given field.

    Parameters
    ----------
    data_source : :class:`yt.data_objects.construction_data_containers.YTQuadTreeProj`
                   or :class:`yt.data_objects.selection_data_containers.YTSlice`
        This is the source to be pixelized, which can be a projection, slice or
        cutting plane.
    bounds : sequence of floats
        Bounds are the min and max in the image plane that we want our
        image to cover.  It's in the order of (xmin, xmax, ymin, ymax),
        where the coordinates are all in the appropriate code units.
    buff_size : sequence of ints
        The size of the image to generate.
    antialias : boolean
        This can be true or false.  It determines whether or not sub-pixel
        rendering is used during data deposition.
    periodic : boolean
        This can be true or false, and governs whether the pixelization
        will span the domain boundaries.

    Examples
    --------
    To make a projection and then several images, you can generate a
    single FRB and then access multiple fields:

    >>> proj = ds.proj(0, ("gas", "density"))
    >>> frb1 = FixedResolutionBuffer(proj, (0.2, 0.3, 0.4, 0.5), (1024, 1024))
    >>> print(frb1[("gas", "density")].max())
    1.0914e-9 g/cm**3
    >>> print(frb1[("gas", "temperature")].max())
    104923.1 K
    """
    _exclude_fields = (
        "pz",
        "pdz",
        "dx",
        "x",
        "y",
        "z",
        "r",
        "dr",
        "phi",
        "dphi",
        "theta",
        "dtheta",
        ("index", "dx"),
        ("index", "x"),
        ("index", "y"),
        ("index", "z"),
        ("index", "r"),
        ("index", "dr"),
        ("index", "phi"),
        ("index", "dphi"),
        ("index", "theta"),
        ("index", "dtheta"),
    )

    def __init__(self, data_source, bounds, buff_size, antialias=True, periodic=False):
        self.data_source = data_source
        self.ds = data_source.ds
        self.bounds = bounds
        self.buff_size = (int(buff_size[0]), int(buff_size[1]))
        self.antialias = antialias
        self.data = {}
        self._filters = []
        self.axis = data_source.axis
        self.periodic = periodic

        ds = getattr(data_source, "ds", None)
        if ds is not None:
            ds.plots.append(weakref.proxy(self))

        # Handle periodicity, just in case
        if self.data_source.axis < 3:
            DLE = self.ds.domain_left_edge
            DRE = self.ds.domain_right_edge
            DD = float(self.periodic) * (DRE - DLE)
            axis = self.data_source.axis
            xax = self.ds.coordinates.x_axis[axis]
            yax = self.ds.coordinates.y_axis[axis]
            self._period = (DD[xax], DD[yax])
            self._edges = ((DLE[xax], DRE[xax]), (DLE[yax], DRE[yax]))

        self.setup_filters()

    def keys(self):
        return self.data.keys()

    def __delitem__(self, item):
        del self.data[item]

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        mylog.info(
            "Making a fixed resolution buffer of (%s) %d by %d",
            item,
            self.buff_size[0],
            self.buff_size[1],
        )
        bounds = []
        for b in self.bounds:
            if hasattr(b, "in_units"):
                b = float(b.in_units("code_length"))
            bounds.append(b)

        buff = self.ds.coordinates.pixelize(
            self.data_source.axis,
            self.data_source,
            item,
            bounds,
            self.buff_size,
            int(self.antialias),
        )

        for name, (args, kwargs) in self._filters:
            buff = filter_registry[name](*args[1:], **kwargs).apply(buff)

        # FIXME FIXME FIXME we shouldn't need to do this for projections
        # but that will require fixing data object access for particle
        # projections
        try:
            if hasattr(item, "name"):
                it = item.name
            else:
                it = item
            units = self.data_source._projected_units[it]
        except (KeyError, AttributeError):
            units = self.data_source[item].units

        ia = ImageArray(buff, units=units, info=self._get_info(item))
        self.data[item] = ia
        return self.data[item]

    def __setitem__(self, item, val):
        self.data[item] = val

    def _get_data_source_fields(self):
        exclude = self.data_source._key_fields + list(self._exclude_fields)
        fields = getattr(self.data_source, "fields", [])
        fields += getattr(self.data_source, "field_data", {}).keys()
        for f in fields:
            if f not in exclude and f[0] not in self.data_source.ds.particle_types:
                self[f]

    def _get_info(self, item):
        info = {}
        ftype, fname = field = self.data_source._determine_fields(item)[0]
        finfo = self.data_source.ds._get_field_info(*field)
        info["data_source"] = self.data_source.__str__()
        info["axis"] = self.data_source.axis
        info["field"] = str(item)
        info["xlim"] = self.bounds[:2]
        info["ylim"] = self.bounds[2:]
        info["length_unit"] = self.data_source.ds.length_unit
        info["length_to_cm"] = info["length_unit"].in_cgs().to_ndarray()
        info["center"] = self.data_source.center

        try:
            info["coord"] = self.data_source.coord
        except AttributeError:
            pass

        try:
            info["weight_field"] = self.data_source.weight_field
        except AttributeError:
            pass

        info["label"] = finfo.get_latex_display_name()

        return info

    def convert_to_pixel(self, coords):
        r"""This function converts coordinates in code-space to pixel-space.

        Parameters
        ----------
        coords : sequence of array_like
            This is (x_coord, y_coord).  Because of the way the math is done,
            these can both be arrays.

        Returns
        -------
        output : sequence of array_like
            This returns px_coord, py_coord

        """
        dpx = (self.bounds[1] - self.bounds[0]) / self.buff_size[0]
        dpy = (self.bounds[3] - self.bounds[2]) / self.buff_size[1]
        px = (coords[0] - self.bounds[0]) / dpx
        py = (coords[1] - self.bounds[2]) / dpy
        return (px, py)

    def convert_distance_x(self, distance):
        r"""This function converts code-space distance into pixel-space
        distance in the x-coordinate.

        Parameters
        ----------
        distance : array_like
            This is x-distance in code-space you would like to convert.

        Returns
        -------
        output : array_like
            The return value is the distance in the y-pixel coordinates.

        """
        dpx = (self.bounds[1] - self.bounds[0]) / self.buff_size[0]
        return distance / dpx

    def convert_distance_y(self, distance):
        r"""This function converts code-space distance into pixel-space
        distance in the y-coordinate.

        Parameters
        ----------
        distance : array_like
            This is y-distance in code-space you would like to convert.

        Returns
        -------
        output : array_like
            The return value is the distance in the x-pixel coordinates.

        """
        dpy = (self.bounds[3] - self.bounds[2]) / self.buff_size[1]
        return distance / dpy

    def set_unit(self, field, unit, equivalency=None, equivalency_kwargs=None):
        """Sets a new unit for the requested field

        parameters
        ----------
        field : string or field tuple
           The name of the field that is to be changed.

        unit : string or Unit object
           The name of the new unit.

        equivalency : string, optional
           If set, the equivalency to use to convert the current units to
           the new requested unit. If None, the unit conversion will be done
           without an equivalency

        equivalency_kwargs : string, optional
           Keyword arguments to be passed to the equivalency. Only used if
           ``equivalency`` is set.
        """
        if equivalency_kwargs is None:
            equivalency_kwargs = {}
        field = self.data_source._determine_fields(field)[0]
        if equivalency is None:
            self[field].convert_to_units(unit)
        else:
            equiv_array = self[field].to_equivalent(
                unit, equivalency, **equivalency_kwargs
            )
            # equiv_array isn't necessarily an ImageArray. This is an issue
            # inherent to the way the unit system handles YTArray
            # subclasses and I don't see how to modify the unit system to
            # fix this. Instead, we paper over this issue and hard code
            # that equiv_array is an ImageArray
            self[field] = ImageArray(
                equiv_array,
                equiv_array.units,
                equiv_array.units.registry,
                self[field].info,
            )

    def export_hdf5(self, filename, fields=None):
        r"""Export a set of fields to a set of HDF5 datasets.

        This function will export any number of fields into datasets in a new
        HDF5 file.

        Parameters
        ----------
        filename : string
            This file will be opened in "append" mode.
        fields : list of strings
            These fields will be pixelized and output.
        """
        if fields is None:
            fields = list(self.data.keys())
        output = h5py.File(filename, mode="a")
        for field in fields:
            output.create_dataset(field, data=self[field])
        output.close()

    def to_fits_data(self, fields=None, other_keys=None, length_unit=None, **kwargs):
        r"""Export the fields in this FixedResolutionBuffer instance
        to a FITSImageData instance.

        This will export a set of FITS images of either the fields specified
        or all the fields already in the object.

        Parameters
        ----------
        fields : list of strings
            These fields will be pixelized and output. If "None", the keys of
            the FRB will be used.
        other_keys : dictionary, optional
            A set of header keys and values to write into the FITS header.
        length_unit : string, optional
            the length units that the coordinates are written in. The default
            is to use the default length unit of the dataset.
        """
        from yt.visualization.fits_image import FITSImageData

        if length_unit is None:
            length_unit = self.ds.length_unit

        if fields is None:
            fields = list(self.data.keys())
        else:
            fields = list(iter_fields(fields))

        if len(fields) == 0:
            raise RuntimeError(
                "No fields to export. Either pass a field or list of fields to "
                "to_fits_data or access a field from the FixedResolutionBuffer "
                "object."
            )

        fid = FITSImageData(self, fields=fields, length_unit=length_unit)
        if other_keys is not None:
            for k, v in other_keys.items():
                fid.update_all_headers(k, v)
        return fid

    def export_dataset(self, fields=None, nprocs=1):
        r"""Export a set of pixelized fields to an in-memory dataset that can be
        analyzed as any other in yt. Unit information and other parameters (e.g.,
        geometry, current_time, etc.) will be taken from the parent dataset.

        Parameters
        ----------
        fields : list of strings, optional
            These fields will be pixelized and output. If "None", the keys of the
            FRB will be used.
        nprocs: integer, optional
            If greater than 1, will create this number of subarrays out of data

        Examples
        --------
        >>> import yt
        >>> ds = yt.load("GasSloshing/sloshing_nomag2_hdf5_plt_cnt_0150")
        >>> slc = ds.slice(2, 0.0)
        >>> frb = slc.to_frb((500.0, "kpc"), 500)
        >>> ds2 = frb.export_dataset(
        ...     fields=[("gas", "density"), ("gas", "temperature")], nprocs=32
        ... )
        """
        nx, ny = self.buff_size
        data = {}
        if fields is None:
            fields = list(self.keys())
        for field in fields:
            arr = self[field]
            data[field] = (arr.d.T.reshape(nx, ny, 1), str(arr.units))
        bounds = [b.in_units("code_length").v for b in self.bounds]
        bbox = np.array([[bounds[0], bounds[1]], [bounds[2], bounds[3]], [0.0, 1.0]])
        return load_uniform_grid(
            data,
            [nx, ny, 1],
            length_unit=self.ds.length_unit,
            bbox=bbox,
            sim_time=self.ds.current_time.in_units("s").v,
            mass_unit=self.ds.mass_unit,
            time_unit=self.ds.time_unit,
            velocity_unit=self.ds.velocity_unit,
            magnetic_unit=self.ds.magnetic_unit,
            periodicity=(False, False, False),
            geometry=self.ds.geometry,
            nprocs=nprocs,
        )

    def save_as_dataset(self, filename=None, fields=None):
        r"""Export a fixed resolution buffer to a reloadable yt dataset.

        This function will take a fixed resolution buffer and output a
        dataset containing either the fields presently existing or fields
        given in the ``fields`` list.  The resulting dataset can be
        reloaded as a yt dataset.

        Parameters
        ----------
        filename : str, optional
            The name of the file to be written.  If None, the name
            will be a combination of the original dataset and the type
            of data container.
        fields : list of strings or tuples, optional
            If this is supplied, it is the list of fields to be saved to
            disk.  If not supplied, all the fields that have been queried
            will be saved.

        Returns
        -------
        filename : str
            The name of the file that has been created.

        Examples
        --------

        >>> import yt
        >>> ds = yt.load("enzo_tiny_cosmology/DD0046/DD0046")
        >>> proj = ds.proj(("gas", "density"), "x", weight_field=("gas", "density"))
        >>> frb = proj.to_frb(1.0, (800, 800))
        >>> fn = frb.save_as_dataset(fields=[("gas", "density")])
        >>> ds2 = yt.load(fn)
        >>> print(ds2.data[("gas", "density")])
        [[  1.25025353e-30   1.25025353e-30   1.25025353e-30 ...,   7.90820691e-31
            7.90820691e-31   7.90820691e-31]
         [  1.25025353e-30   1.25025353e-30   1.25025353e-30 ...,   7.90820691e-31
            7.90820691e-31   7.90820691e-31]
         [  1.25025353e-30   1.25025353e-30   1.25025353e-30 ...,   7.90820691e-31
            7.90820691e-31   7.90820691e-31]
         ...,
         [  1.55834239e-30   1.55834239e-30   1.55834239e-30 ...,   8.51353199e-31
            8.51353199e-31   8.51353199e-31]
         [  1.55834239e-30   1.55834239e-30   1.55834239e-30 ...,   8.51353199e-31
            8.51353199e-31   8.51353199e-31]
         [  1.55834239e-30   1.55834239e-30   1.55834239e-30 ...,   8.51353199e-31
            8.51353199e-31   8.51353199e-31]] g/cm**3

        """

        keyword = f"{str(self.ds)}_{self.data_source._type_name}_frb"
        filename = get_output_filename(filename, keyword, ".h5")

        data = {}
        if fields is not None:
            for f in self.data_source._determine_fields(fields):
                data[f] = self[f]
        else:
            data.update(self.data)

        ftypes = {field: "grid" for field in data}
        extra_attrs = {
            arg: getattr(self.data_source, arg, None)
            for arg in self.data_source._con_args + self.data_source._tds_attrs
        }
        extra_attrs["con_args"] = self.data_source._con_args
        extra_attrs["left_edge"] = self.ds.arr([self.bounds[0], self.bounds[2]])
        extra_attrs["right_edge"] = self.ds.arr([self.bounds[1], self.bounds[3]])
        # The data dimensions are [NY, NX] but buff_size is [NX, NY].
        extra_attrs["ActiveDimensions"] = self.buff_size[::-1]
        extra_attrs["level"] = 0
        extra_attrs["data_type"] = "yt_frb"
        extra_attrs["container_type"] = self.data_source._type_name
        extra_attrs["dimensionality"] = self.data_source._dimensionality
        save_as_dataset(
            self.ds, filename, data, field_types=ftypes, extra_attrs=extra_attrs
        )

        return filename

    @property
    def limits(self):
        rv = dict(x=None, y=None, z=None)
        xax = self.ds.coordinates.x_axis[self.axis]
        yax = self.ds.coordinates.y_axis[self.axis]
        xn = self.ds.coordinates.axis_name[xax]
        yn = self.ds.coordinates.axis_name[yax]
        rv[xn] = (self.bounds[0], self.bounds[1])
        rv[yn] = (self.bounds[2], self.bounds[3])
        return rv

    def setup_filters(self):
        ignored = ["FixedResolutionBufferFilter"]
        for key in filter_registry:
            if key in ignored:
                continue
            filtername = filter_registry[key]._filter_name

            # We need to wrap to create a closure so that
            # FilterMaker is bound to the wrapped method.
            def closure():
                FilterMaker = filter_registry[key]

                @wraps(FilterMaker)
                def method(*args, **kwargs):
                    # We need to also do it here as "invalidate_plot"
                    # and "apply_callback" require the functions'
                    # __name__ in order to work properly
                    @wraps(FilterMaker)
                    def cb(self, *a, **kwa):
                        # We construct the callback method
                        # skipping self
                        return FilterMaker(*a, **kwa)

                    # Create callback
                    cb = apply_filter(cb)

                    return cb(self, *args, **kwargs)

                return method

            self.__dict__["apply_" + filtername] = closure()


class CylindricalFixedResolutionBuffer(FixedResolutionBuffer):
    """
    This object is a subclass of
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer`
    that supports non-aligned input data objects, primarily cutting planes.
    """

    def __init__(self, data_source, radius, buff_size, antialias=True):

        self.data_source = data_source
        self.ds = data_source.ds
        self.radius = radius
        self.buff_size = buff_size
        self.antialias = antialias
        self.data = {}

        ds = getattr(data_source, "ds", None)
        if ds is not None:
            ds.plots.append(weakref.proxy(self))

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        buff = np.zeros(self.buff_size, dtype="f8")
        pixelize_cylinder(
            buff,
            self.data_source["r"],
            self.data_source["dr"],
            self.data_source["theta"],
            self.data_source["dtheta"],
            self.data_source[item].astype("float64"),
            self.radius,
        )
        self[item] = buff
        return buff


class OffAxisProjectionFixedResolutionBuffer(FixedResolutionBuffer):
    """
    This object is a subclass of
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer`
    that supports off axis projections.  This calls the volume renderer.
    """

    def __init__(self, data_source, bounds, buff_size, antialias=True, periodic=False):
        self.data = {}
        FixedResolutionBuffer.__init__(
            self, data_source, bounds, buff_size, antialias, periodic
        )

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]
        mylog.info(
            "Making a fixed resolution buffer of (%s) %d by %d",
            item,
            self.buff_size[0],
            self.buff_size[1],
        )
        dd = self.data_source
        width = self.ds.arr(
            (
                self.bounds[1] - self.bounds[0],
                self.bounds[3] - self.bounds[2],
                self.bounds[5] - self.bounds[4],
            )
        )
        buff = off_axis_projection(
            dd.dd,
            dd.center,
            dd.normal_vector,
            width,
            self.buff_size,
            item,
            weight=dd.weight_field,
            volume=dd.volume,
            no_ghost=dd.no_ghost,
            interpolated=dd.interpolated,
            north_vector=dd.north_vector,
            method=dd.method,
        )
        ia = ImageArray(buff.swapaxes(0, 1), info=self._get_info(item))
        self[item] = ia
        return ia


class ParticleImageBuffer(FixedResolutionBuffer):
    """

    This object is a subclass of
    :class:`yt.visualization.fixed_resolution.FixedResolutionBuffer`
    that supports particle plots. It splats points onto an image
    buffer.

    """

    def __init__(self, data_source, bounds, buff_size, antialias=True, periodic=False):
        self.data = {}
        FixedResolutionBuffer.__init__(
            self, data_source, bounds, buff_size, antialias, periodic
        )

        # set up the axis field names
        axis = self.axis
        self.xax = self.ds.coordinates.x_axis[axis]
        self.yax = self.ds.coordinates.y_axis[axis]
        ax_field_template = "particle_position_%s"
        self.x_field = ax_field_template % self.ds.coordinates.axis_name[self.xax]
        self.y_field = ax_field_template % self.ds.coordinates.axis_name[self.yax]

    def __getitem__(self, item):
        if item in self.data:
            return self.data[item]

        deposition = self.data_source.deposition
        density = self.data_source.density

        mylog.info(
            "Splatting (%s) onto a %d by %d mesh using method '%s'",
            item,
            self.buff_size[0],
            self.buff_size[1],
            deposition,
        )

        bounds = []
        for b in self.bounds:
            if hasattr(b, "in_units"):
                b = float(b.in_units("code_length"))
            bounds.append(b)

        ftype = item[0]
        x_data = self.data_source.dd[ftype, self.x_field]
        y_data = self.data_source.dd[ftype, self.y_field]
        data = self.data_source.dd[item]

        # handle periodicity
        dx = x_data.in_units("code_length").d - bounds[0]
        dy = y_data.in_units("code_length").d - bounds[2]
        if self.periodic:
            dx %= float(self._period[0].in_units("code_length"))
            dy %= float(self._period[1].in_units("code_length"))

        # convert to pixels
        px = dx / (bounds[1] - bounds[0])
        py = dy / (bounds[3] - bounds[2])

        # select only the particles that will actually show up in the image
        mask = np.logical_and(
            np.logical_and(px >= 0.0, px <= 1.0), np.logical_and(py >= 0.0, py <= 1.0)
        )

        weight_field = self.data_source.weight_field
        if weight_field is None:
            weight_data = np.ones_like(data.v)
        else:
            weight_data = self.data_source.dd[weight_field]
        splat_vals = weight_data[mask] * data[mask]

        x_bin_edges = np.linspace(0.0, 1.0, self.buff_size[0] + 1)
        y_bin_edges = np.linspace(0.0, 1.0, self.buff_size[1] + 1)

        # splat particles
        buff = np.zeros(self.buff_size)
        buff_mask = np.zeros_like(buff, dtype="uint8")
        if deposition == "ngp":
            add_points_to_greyscale_image(
                buff, buff_mask, px[mask], py[mask], splat_vals
            )
        elif deposition == "cic":
            CICDeposit_2(
                py[mask],
                px[mask],
                splat_vals,
                mask.sum(),
                buff,
                buff_mask,
                x_bin_edges,
                y_bin_edges,
            )
        else:
            raise ValueError(f"Received unknown deposition method '{deposition}'")

        # remove values in no-particle region
        buff[buff_mask == 0] = np.nan

        # Normalize by the surface area of the pixel or volume of pencil if
        # requested
        info = self._get_info(item)
        if density:
            width = self.data_source.width
            norm = width[self.xax] * width[self.yax] / np.prod(self.buff_size)
            norm = norm.in_base()
            buff /= norm.v
            units = data.units / norm.units
            info["label"] = "%s $\\rm{Density}$" % info["label"]
        else:
            units = data.units

        ia = ImageArray(buff, units=units, info=info)

        # divide by the weight_field, if needed
        if weight_field is not None:
            weight_buff = np.zeros(self.buff_size)
            weight_buff_mask = np.zeros(self.buff_size, dtype="uint8")
            if deposition == "ngp":
                add_points_to_greyscale_image(
                    weight_buff, weight_buff_mask, px[mask], py[mask], weight_data[mask]
                )
            elif deposition == "cic":
                CICDeposit_2(
                    py[mask],
                    px[mask],
                    weight_data[mask],
                    mask.sum(),
                    weight_buff,
                    weight_buff_mask,
                    y_bin_edges,
                    x_bin_edges,
                )
            weight_array = ImageArray(
                weight_buff, units=weight_data.units, info=self._get_info(item)
            )
            # remove values in no-particle region
            weight_buff[weight_buff_mask == 0] = np.nan
            locs = np.where(weight_array > 0)
            ia[locs] /= weight_array[locs]

        self.data[item] = ia
        return self.data[item]

    # over-ride the base class version, since we don't want to exclude
    # particle fields
    def _get_data_source_fields(self):
        exclude = self.data_source._key_fields + list(self._exclude_fields)
        fields = getattr(self.data_source, "fields", [])
        fields += getattr(self.data_source, "field_data", {}).keys()
        for f in fields:
            if f not in exclude:
                self[f]



import os
import weakref

import numpy as np

from yt.data_objects.index_subobjects.unstructured_mesh import SemiStructuredMesh
from yt.data_objects.static_output import Dataset
from yt.funcs import setdefaultattr
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.utilities.file_handler import HDF5FileHandler
from yt.utilities.on_demand_imports import _h5py as h5py

from .fields import MoabFieldInfo, PyneFieldInfo


class MoabHex8Mesh(SemiStructuredMesh):
    _connectivity_length = 8
    _index_offset = 1


class MoabHex8Hierarchy(UnstructuredIndex):
    def __init__(self, ds, dataset_type="h5m"):
        self.dataset = weakref.proxy(ds)
        self.dataset_type = dataset_type
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.path.dirname(self.index_filename)
        self._fhandle = h5py.File(self.index_filename, mode="r")

        UnstructuredIndex.__init__(self, ds, dataset_type)

        self._fhandle.close()

    def _initialize_mesh(self):
        con = self._fhandle["/tstt/elements/Hex8/connectivity"][:]
        con = np.asarray(con, dtype="int64")
        coords = self._fhandle["/tstt/nodes/coordinates"][:]
        coords = np.asarray(coords, dtype="float64")
        self.meshes = [MoabHex8Mesh(0, self.index_filename, con, coords, self)]

    def _detect_output_fields(self):
        self.field_list = [
            ("moab", f) for f in self._fhandle["/tstt/elements/Hex8/tags"].keys()
        ]

    def _count_grids(self):
        self.num_grids = 1


class MoabHex8Dataset(Dataset):
    _index_class = MoabHex8Hierarchy
    _field_info_class = MoabFieldInfo
    periodicity = (False, False, False)

    def __init__(
        self,
        filename,
        dataset_type="moab_hex8",
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.fluid_types += ("moab",)
        Dataset.__init__(
            self,
            filename,
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = storage_filename
        self.filename = filename
        self._handle = HDF5FileHandler(filename)

    def _set_code_unit_attributes(self):
        # Almost everything is regarded as dimensionless in MOAB, so these will
        # not be used very much or at all.
        setdefaultattr(self, "length_unit", self.quan(1.0, "cm"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "g"))

    def _parse_parameter_file(self):
        self._handle = h5py.File(self.parameter_filename, mode="r")
        coords = self._handle["/tstt/nodes/coordinates"]
        self.domain_left_edge = coords[0]
        self.domain_right_edge = coords[-1]
        self.domain_dimensions = self.domain_right_edge - self.domain_left_edge
        self.refine_by = 2
        self.dimensionality = len(self.domain_dimensions)
        self.current_time = 0.0
        self.unique_identifier = self.parameter_filename
        self.cosmological_simulation = False
        self.num_ghost_zones = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0
        self.cosmological_simulation = 0

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return filename.endswith(".h5m")

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]


class PyneHex8Mesh(SemiStructuredMesh):
    _connectivity_length = 8
    _index_offset = 0


class PyneMeshHex8Hierarchy(UnstructuredIndex):
    def __init__(self, ds, dataset_type="moab_hex8_pyne"):
        self.dataset = weakref.proxy(ds)
        self.dataset_type = dataset_type
        self.index_filename = self.dataset.parameter_filename
        self.directory = os.getcwd()
        self.pyne_mesh = ds.pyne_mesh

        super().__init__(ds, dataset_type)

    def _initialize_mesh(self):
        from pymoab import types

        ents = list(self.pyne_mesh.structured_iterate_vertex())
        coords = self.pyne_mesh.mesh.get_coords(ents).astype("float64")
        coords = coords.reshape(len(coords) // 3, 3)
        hexes = self.pyne_mesh.mesh.get_entities_by_type(0, types.MBHEX)
        vind = []
        for h in hexes:
            vind.append(
                self.pyne_mesh.mesh.get_adjacencies(
                    h, 0, create_if_missing=True, op_type=types.UNION
                )
            )
        vind = np.asarray(vind, dtype=np.int64)
        vind = vind.reshape(len(vind) // 8, 8)
        self.meshes = [PyneHex8Mesh(0, self.index_filename, vind, coords, self)]

    def _detect_output_fields(self):
        self.field_list = [("pyne", f) for f in self.pyne_mesh.tags.keys()]

    def _count_grids(self):
        self.num_grids = 1


class PyneMoabHex8Dataset(Dataset):
    _index_class = PyneMeshHex8Hierarchy
    _fieldinfo_fallback = MoabFieldInfo
    _field_info_class = PyneFieldInfo
    periodicity = (False, False, False)

    def __init__(
        self,
        pyne_mesh,
        dataset_type="moab_hex8_pyne",
        storage_filename=None,
        units_override=None,
        unit_system="cgs",
    ):
        self.fluid_types += ("pyne",)
        filename = f"pyne_mesh_{id(pyne_mesh)}"
        self.pyne_mesh = pyne_mesh
        Dataset.__init__(
            self,
            str(filename),
            dataset_type,
            units_override=units_override,
            unit_system=unit_system,
        )
        self.storage_filename = storage_filename
        self.filename = filename

    def _set_code_unit_attributes(self):
        # Almost everything is regarded as dimensionless in MOAB, so these will
        # not be used very much or at all.
        setdefaultattr(self, "length_unit", self.quan(1.0, "cm"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "g"))

    def _parse_parameter_file(self):
        ents = list(self.pyne_mesh.structured_iterate_vertex())
        coords = self.pyne_mesh.mesh.get_coords(ents)
        self.domain_left_edge = coords[0:3]
        self.domain_right_edge = coords[-3:]
        self.domain_dimensions = self.domain_right_edge - self.domain_left_edge
        self.refine_by = 2
        self.dimensionality = len(self.domain_dimensions)
        self.current_time = 0.0
        self.unique_identifier = self.parameter_filename
        self.cosmological_simulation = False
        self.num_ghost_zones = 0
        self.current_redshift = 0.0
        self.omega_lambda = 0.0
        self.omega_matter = 0.0
        self.hubble_constant = 0.0
        self.cosmological_simulation = 0

    @classmethod
    def _is_valid(cls, filename, *args, **kwargs):
        return False

    def __str__(self):
        return self.basename.rsplit(".", 1)[0]



import matplotlib
from matplotlib.backends.backend_agg import FigureCanvasAgg
from matplotlib.backends.backend_pdf import FigureCanvasPdf
from matplotlib.backends.backend_ps import FigureCanvasPS
from matplotlib.backends.backend_svg import FigureCanvasSVG

matplotlib.rc("contour", negative_linestyle="solid")



import warnings


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning, adapted from NumPy

    The nose runner does not show users DeprecationWarning.
    This ensures that a deprecation warning is visible to users
    if that is desired.
    """

    # this class becomes useless after the tests are migrated from nose to pytest

    pass


def issue_deprecation_warning(msg, *, removal, since=None, stacklevel=3):
    """
    Parameters
    ----------
    msg : str
        A text message explaining that the code surrounding the call to this function is
        deprecated, and what should be changed on the user side to avoid it.

    since and removal: str version numbers, indicating the anticipated removal date

    Crucial note:
    beware that `removal` is required (it doesn't have a default value). This is
    vital since deprecated code is typically untested and not specifying a required
    keyword argument will turn the warning into a TypeError.
    What it gets us however is that the release manager will know for a fact wether it
    is safe to remove a feature at any given point, and users have a better idea when
    their code will become incompatible.

    `since` is optional only because it was introduced in the 4.0.0 release, and it
    should become mandatory in the future.
    Both `since` and `removal` are keyword-only arguments so that their order can be
    swapped in the future without introducing bugs.

    Examples
    --------
    >>> issue_deprecation_warning(
    ...     "This code is deprecated.", since="4.0.0", removal="4.2.0"
    ... )
    """
    msg += "\n"
    if since is not None:
        msg += f"Deprecated since v{since}. "

    msg += f"This feature will be removed in v{removal}"
    warnings.warn(msg, VisibleDeprecationWarning, stacklevel=stacklevel)



import numpy as np
import pyx
from matplotlib import cm, pyplot as plt

from yt._maintenance.deprecation import issue_deprecation_warning
from yt.config import ytcfg
from yt.units.unit_object import Unit
from yt.units.yt_array import YTQuantity
from yt.utilities.logger import ytLogger as mylog

from .plot_window import PlotWindow
from .profile_plotter import PhasePlot, ProfilePlot


def convert_frac_to_tex(string):
    frac_pos = string.find(r"\frac")
    result = string[frac_pos + 5 :]
    level = [0] * len(result)
    clevel = 0
    for i in range(len(result)):
        if result[i] == "{":
            clevel += 1
        elif result[i] == "}":
            clevel -= 1
        level[i] = clevel
    div_pos = level.index(0)
    end_pos = level.index(0, div_pos + 1)
    result = (
        r"${"
        + result[: div_pos + 1]
        + r"\over"
        + result[div_pos + 1 : end_pos]
        + r"}$"
        + result[end_pos:]
    )
    result = result.replace(r"\ ", r"\;")
    return result


def pyxize_label(string):
    frac_pos = string.find(r"\frac")
    if frac_pos >= 0:
        pre = string[:frac_pos]
        result = pre + convert_frac_to_tex(string)
    else:
        result = string
    result = result.replace("$", "")
    result = r"$" + result + r"$"
    return result


class DualEPS:
    def __init__(self, figsize=(12, 12)):
        r"""Initializes the DualEPS class to which we can progressively add layers
        of vector graphics and compressed bitmaps.

        Parameters
        ----------
        figsize : tuple of floats
            The width and height of a single figure in centimeters.
        """
        pyx.unit.set(xscale=1.4)
        self.figsize = figsize
        self.canvas = None
        self.colormaps = None
        self.field = None
        self.axes_drawn = False

    def hello_world(self):
        r"""A simple test."""
        if self.canvas is None:
            self.canvas = pyx.canvas.canvas()
        p = pyx.path.line(0, 0, 1, 1)
        self.canvas.stroke(p)
        self.canvas.text(0, 0, "Hello world.")

    # =============================================================================

    def return_field(self, plot):
        if isinstance(plot, (PlotWindow, PhasePlot)):
            return list(plot.plots.keys())[0]
        else:
            return None

    # =============================================================================

    def axis_box(
        self,
        xrange=(0, 1),
        yrange=(0, 1),
        xlabel="",
        ylabel="",
        xlog=False,
        ylog=False,
        xdata=None,
        ydata=None,
        tickcolor=None,
        bare_axes=False,
        pos=(0, 0),
        xaxis_side=0,
        yaxis_side=0,
        size=None,
    ):
        r"""Draws an axis box in the figure.

        Parameters
        ----------
        xrange : tuple of floats
            The min and max of the x-axis
        yrange : tuple of floats
            The min and max of the y-axis
        xlabel : string
            Label for the x-axis
        ylabel : string
            Label for the y-axis
        xlog : boolean
            Flag to use a logarithmic x-axis
        ylog : boolean
            Flag to use a logarithmic y-axis
        tickcolor : `pyx.color.*.*`
            Color for the tickmarks.  Example: pyx.color.cmyk.black
        bare_axes : boolean
            Set to true to have no annotations or tick marks on all of the
            axes.
        pos : tuple of floats
            (x,y) position in centimeters of the origin in the figure
        xaxis_side : integer
            Set to 0 for the x-axis annotations to be on the left.  Set
            to 1 to print them on the right side.
        yaxis_side : integer
            Set to 0 for the y-axis annotations to be on the bottom.  Set
            to 1 to print them on the top.
        size : tuple of floats
            Size of axis box in units of figsize

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.save_fig()
        """

        if isinstance(xrange[0], YTQuantity):
            xrange = (xrange[0].value, xrange[1].value)
        if isinstance(yrange[0], YTQuantity):
            yrange = (yrange[0].value, yrange[1].value)
        if tickcolor is None:
            c1 = pyx.graph.axis.painter.regular(tickattrs=[pyx.color.cmyk.black])
            c2 = pyx.graph.axis.painter.regular(
                tickattrs=[pyx.color.cmyk.black], labelattrs=None
            )
        else:
            c1 = pyx.graph.axis.painter.regular(tickattrs=[tickcolor])
            c2 = pyx.graph.axis.painter.regular(tickattrs=[tickcolor], labelattrs=None)

        if size is None:
            psize = self.figsize
        else:
            psize = (size[0] * self.figsize[0], size[1] * self.figsize[1])

        xticklabels = True
        yticklabels = True
        if xaxis_side == 0:
            xleftlabel = xlabel
            xrightlabel = ""
            c1x = c1
            c2x = c2
        elif xaxis_side == 1:
            xleftlabel = ""
            xrightlabel = xlabel
            c1x = c2
            c2x = c1
        else:
            xticklabels = False
            xleftlabel = ""
            xrightlabel = ""
            c1x = c1
            c2x = c2
        if yaxis_side == 0:
            yleftlabel = ylabel
            yrightlabel = ""
            c1y = c1
            c2y = c2
        elif yaxis_side == 1:
            yleftlabel = ""
            yrightlabel = ylabel
            c1y = c2
            c2y = c1
        else:
            yticklabels = False
            yleftlabel = ""
            yrightlabel = ""
            c1y = c1
            c2y = c2

        if xlog:
            if xticklabels:
                xaxis = pyx.graph.axis.log(
                    min=xrange[0], max=xrange[1], title=xleftlabel, painter=c1x
                )
                xaxis2 = pyx.graph.axis.log(
                    min=xrange[0], max=xrange[1], title=xrightlabel, painter=c2x
                )
            else:
                xaxis = pyx.graph.axis.log(
                    min=xrange[0],
                    max=xrange[1],
                    title=xleftlabel,
                    painter=c1x,
                    parter=None,
                )
                xaxis2 = pyx.graph.axis.log(
                    min=xrange[0],
                    max=xrange[1],
                    title=xrightlabel,
                    painter=c2x,
                    parter=None,
                )
        else:
            if xticklabels:
                xaxis = pyx.graph.axis.lin(
                    min=xrange[0], max=xrange[1], title=xleftlabel, painter=c1x
                )
                xaxis2 = pyx.graph.axis.lin(
                    min=xrange[0], max=xrange[1], title=xrightlabel, painter=c2x
                )
            else:
                xaxis = pyx.graph.axis.lin(
                    min=xrange[0],
                    max=xrange[1],
                    title=xleftlabel,
                    painter=c1x,
                    parter=None,
                )
                xaxis2 = pyx.graph.axis.lin(
                    min=xrange[0],
                    max=xrange[1],
                    title=xrightlabel,
                    painter=c2x,
                    parter=None,
                )
        if ylog:
            if yticklabels:
                yaxis = pyx.graph.axis.log(
                    min=yrange[0], max=yrange[1], title=yleftlabel, painter=c1y
                )
                yaxis2 = pyx.graph.axis.log(
                    min=yrange[0], max=yrange[1], title=yrightlabel, painter=c2y
                )
            else:
                yaxis = pyx.graph.axis.log(
                    min=yrange[0],
                    max=yrange[1],
                    title=yleftlabel,
                    painter=c1y,
                    parter=None,
                )
                yaxis2 = pyx.graph.axis.log(
                    min=yrange[0],
                    max=yrange[1],
                    title=yrightlabel,
                    painter=c2y,
                    parter=None,
                )
        else:
            if yticklabels:
                yaxis = pyx.graph.axis.lin(
                    min=yrange[0], max=yrange[1], title=yleftlabel, painter=c1y
                )
                yaxis2 = pyx.graph.axis.lin(
                    min=yrange[0], max=yrange[1], title=yrightlabel, painter=c2y
                )
            else:
                yaxis = pyx.graph.axis.lin(
                    min=yrange[0],
                    max=yrange[1],
                    title=yleftlabel,
                    painter=c1y,
                    parter=None,
                )
                yaxis2 = pyx.graph.axis.lin(
                    min=yrange[0],
                    max=yrange[1],
                    title=yrightlabel,
                    painter=c2y,
                    parter=None,
                )

        if bare_axes:
            if ylog:
                yaxis = pyx.graph.axis.log(
                    min=yrange[0], max=yrange[1], title=yleftlabel, parter=None
                )
                yaxis2 = pyx.graph.axis.log(
                    min=yrange[0], max=yrange[1], title=yrightlabel, parter=None
                )
            else:
                yaxis = pyx.graph.axis.lin(
                    min=yrange[0], max=yrange[1], title=yleftlabel, parter=None
                )
                yaxis2 = pyx.graph.axis.lin(
                    min=yrange[0], max=yrange[1], title=yrightlabel, parter=None
                )
            if xlog:
                xaxis = pyx.graph.axis.log(
                    min=xrange[0], max=xrange[1], title=xleftlabel, parter=None
                )
                xaxis2 = pyx.graph.axis.log(
                    min=xrange[0], max=xrange[1], title=xrightlabel, parter=None
                )
            else:
                xaxis = pyx.graph.axis.lin(
                    min=xrange[0], max=xrange[1], title=xleftlabel, parter=None
                )
                xaxis2 = pyx.graph.axis.lin(
                    min=xrange[0], max=xrange[1], title=xrightlabel, parter=None
                )

        blank_data = pyx.graph.data.points([(-1e20, -1e20), (-1e19, -1e19)], x=1, y=2)
        if self.canvas is None:
            self.canvas = pyx.graph.graphxy(
                width=psize[0],
                height=psize[1],
                x=xaxis,
                y=yaxis,
                x2=xaxis2,
                y2=yaxis2,
                xpos=pos[0],
                ypos=pos[1],
            )
            if xdata is None:
                self.canvas.plot(blank_data)
            else:
                data = pyx.graph.data.points(np.array([xdata, ydata]).T, x=1, y=2)
                self.canvas.plot(
                    data, [pyx.graph.style.line([pyx.style.linewidth.Thick])]
                )
        else:
            plot = pyx.graph.graphxy(
                width=psize[0],
                height=psize[1],
                x=xaxis,
                y=yaxis,
                x2=xaxis2,
                y2=yaxis2,
                xpos=pos[0],
                ypos=pos[1],
            )
            if xdata is None:
                plot.plot(blank_data)
            else:
                data = pyx.graph.data.points(np.array([xdata, ydata]).T, x=1, y=2)
                plot.plot(data, [pyx.graph.style.line([pyx.style.linewidth.Thick])])
            self.canvas.insert(plot)
        self.axes_drawn = True

    # =============================================================================

    def axis_box_yt(
        self,
        plot,
        units=None,
        bare_axes=False,
        tickcolor=None,
        xlabel=None,
        ylabel=None,
        **kwargs,
    ):
        r"""Wrapper around DualEPS.axis_box to automatically fill in the
        axis ranges and labels from a yt plot.

        This also accepts any parameters that DualEPS.axis_box takes.

        Parameters
        ----------
        plot : `yt.visualization.plot_window.PlotWindow`
            yt plot on which the axes are based.
        units : string
            Unit description that overrides yt's unit description.  Only
            affects the axis label.
        bare_axes : boolean
            Set to true to have no annotations or tick marks on all of the
            axes.

        Examples
        --------
        >>> p = SlicePlot(ds, 0, "density")
        >>> d = DualEPS()
        >>> d.axis_box_yt(p)
        >>> d.save_fig()
        """

        if isinstance(plot, (PlotWindow, PhasePlot)):
            plot.refresh()
        if isinstance(plot, PlotWindow):
            data = plot.frb
            width = plot.width[0]
            if units is None:
                units = plot.ds.get_smallest_appropriate_unit(width)
            width = width.in_units(str(units))
            xc = 0.5 * (plot.xlim[0] + plot.xlim[1])
            yc = 0.5 * (plot.ylim[0] + plot.ylim[1])
            _xrange = [(plot.xlim[i] - xc).in_units(units) for i in (0, 1)]
            _yrange = [(plot.ylim[i] - yc).in_units(units) for i in (0, 1)]
            _xlog = False
            _ylog = False
            if bare_axes:
                _xlabel = ""
                _ylabel = ""
            else:
                if xlabel is not None:
                    _xlabel = xlabel
                else:
                    if data.axis != 4:
                        xi = plot.ds.coordinates.x_axis[data.axis]
                        x_name = plot.ds.coordinates.axis_name[xi]
                        _xlabel = f"{x_name} ({units})"
                    else:
                        _xlabel = f"x ({units})"
                if ylabel is not None:
                    _ylabel = ylabel
                else:
                    if data.axis != 4:
                        yi = plot.ds.coordinates.y_axis[data.axis]
                        y_name = plot.ds.coordinates.axis_name[yi]
                        _ylabel = f"{y_name} ({units})"
                    else:
                        _ylabel = f"y ({units})"
            if tickcolor is None:
                _tickcolor = pyx.color.cmyk.white
        elif isinstance(plot, ProfilePlot):
            subplot = plot.axes.values()[0]
            # limits for axes
            xlimits = subplot.get_xlim()
            _xrange = (
                YTQuantity(xlimits[0], "m"),
                YTQuantity(xlimits[1], "m"),
            )  # unit hardcoded but afaik it is not used anywhere so it doesn't matter
            if list(plot.axes.ylim.values())[0][0] is None:
                ylimits = subplot.get_ylim()
            else:
                ylimits = list(plot.axes.ylim.values())[0]
            _yrange = (
                YTQuantity(ylimits[0], "m"),
                YTQuantity(ylimits[1], "m"),
            )  # unit hardcoded but afaik it is not used anywhere so it doesn't matter
            # axis labels
            xaxis = subplot.xaxis
            _xlabel = pyxize_label(xaxis.label.get_text())
            yaxis = subplot.yaxis
            _ylabel = pyxize_label(yaxis.label.get_text())
            # set log if necessary
            if subplot.get_xscale() == "log":
                _xlog = True
            else:
                _xlog = False
            if subplot.get_yscale() == "log":
                _ylog = True
            else:
                _ylog = False
            _tickcolor = None
        elif isinstance(plot, PhasePlot):
            k = list(plot.plots.keys())[0]
            _xrange = plot[k].axes.get_xlim()
            _yrange = plot[k].axes.get_ylim()
            _xlog = plot.profile.x_log
            _ylog = plot.profile.y_log
            if bare_axes:
                _xlabel = ""
                _ylabel = ""
            else:
                if xlabel is not None:
                    _xlabel = xlabel
                else:
                    _xlabel = plot[k].axes.get_xlabel()
                if ylabel is not None:
                    _ylabel = ylabel
                else:
                    _ylabel = plot[k].axes.get_ylabel()
                _xlabel = pyxize_label(_xlabel)
                _ylabel = pyxize_label(_ylabel)
            if tickcolor is None:
                _tickcolor = None
        elif isinstance(plot, np.ndarray):
            ax = plt.gca()
            _xrange = ax.get_xlim()
            _yrange = ax.get_ylim()
            _xlog = False
            _ylog = False
            if bare_axes:
                _xlabel = ""
                _ylabel = ""
            else:
                if xlabel is not None:
                    _xlabel = xlabel
                else:
                    _xlabel = ax.get_xlabel()
                if ylabel is not None:
                    _ylabel = ylabel
                else:
                    _ylabel = ax.get_ylabel()
            if tickcolor is None:
                _tickcolor = None
        else:
            _xrange = plot._axes.get_xlim()
            _yrange = plot._axes.get_ylim()
            _xlog = plot._log_x
            _ylog = plot._log_y
            if bare_axes:
                _xlabel = ""
                _ylabel = ""
            else:
                if xlabel is not None:
                    _xlabel = xlabel
                else:
                    _xlabel = plot._x_label
                if ylabel is not None:
                    _ylabel = ylabel
                else:
                    _ylabel = plot._y_label
            if tickcolor is None:
                _tickcolor = None
        if tickcolor is not None:
            _tickcolor = tickcolor
        self.axis_box(
            xrange=_xrange,
            yrange=_yrange,
            xlabel=_xlabel,
            ylabel=_ylabel,
            tickcolor=_tickcolor,
            xlog=_xlog,
            ylog=_ylog,
            bare_axes=bare_axes,
            **kwargs,
        )

    # =============================================================================

    def insert_image(self, filename, pos=(0, 0), size=None):
        r"""Inserts a JPEG file in the figure.

        Parameters
        ----------
        filename : string
            Name of the JPEG file
        pos : tuple of floats
            Position of the origin of the image in centimeters
        size : tuple of flots
            Size of image in units of figsize

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("image.jpg")
        >>> d.save_fig()
        """
        if size is not None:
            width = size[0] * self.figsize[0]
            height = size[1] * self.figsize[1]
        else:
            width = self.figsize[0]
            height = self.figsize[1]
        image = pyx.bitmap.jpegimage(filename)
        if self.canvas is None:
            self.canvas = pyx.canvas.canvas()
        self.canvas.insert(
            pyx.bitmap.bitmap(
                pos[0], pos[1], image, compressmode=None, width=width, height=height
            )
        )

    # =============================================================================

    def insert_image_yt(self, plot, field=None, pos=(0, 0), scale=1.0):
        r"""Inserts a bitmap taken from a yt plot.

        Parameters
        ----------
        plot : `yt.visualization.plot_window.PlotWindow`
            yt plot that provides the image
        pos : tuple of floats
            Position of the origin of the image in centimeters.

        Examples
        --------
        >>> p = SlicePlot(ds, 0, "density")
        >>> d = DualEPS()
        >>> d.axis_box_yt(p)
        >>> d.insert_image_yt(p)
        >>> d.save_fig()

        Notes
        -----
        For best results, set use_colorbar=False when creating the yt
        image.
        """
        from ._mpl_imports import FigureCanvasAgg

        # We need to remove the colorbar (if necessary), remove the
        # axes, and resize the figure to span the entire figure
        force_square = False
        if self.canvas is None:
            self.canvas = pyx.canvas.canvas()
        if isinstance(plot, (PlotWindow, PhasePlot)):
            if field is None:
                self.field = list(plot.plots.keys())[0]
                mylog.warning(
                    "No field specified.  Choosing first field (%s)", self.field
                )
            else:
                self.field = plot.data_source._determine_fields(field)[0]
            if self.field not in plot.plots.keys():
                raise RuntimeError("Field '%s' does not exist!", self.field)
            if isinstance(plot, PlotWindow):
                plot.hide_colorbar()
                plot.hide_axes()
            else:
                plot.plots[self.field]._toggle_axes(False)
                plot.plots[self.field]._toggle_colorbar(False)
            plot.refresh()
            _p1 = plot.plots[self.field].figure
            force_square = True
        elif isinstance(plot, ProfilePlot):
            _p1 = plot.figures.items()[0][1]
        elif isinstance(plot, np.ndarray):
            plt.figure()
            iplot = plt.figimage(plot)
            _p1 = iplot.figure
            _p1.set_size_inches(self.figsize[0], self.figsize[1])
            ax = plt.gca()
            _p1.add_axes(ax)
        else:
            raise RuntimeError("Unknown plot type")

        _p1.axes[0].set_position([0, 0, 1, 1])  # rescale figure
        _p1.set_facecolor("w")  # set background color
        figure_canvas = FigureCanvasAgg(_p1)
        figure_canvas.draw()
        size = (_p1.get_size_inches() * _p1.dpi).astype("int")

        # Account for non-square images after removing the colorbar.
        scale *= 1.0 - 1.0 / (_p1.dpi * self.figsize[0])
        if force_square:
            yscale = scale * float(size[1]) / float(size[0])
        else:
            yscale = scale
        image = pyx.bitmap.image(size[0], size[1], "RGB", figure_canvas.tostring_rgb())
        self.canvas.insert(
            pyx.bitmap.bitmap(
                pos[0],
                pos[1],
                image,
                width=scale * self.figsize[0],
                height=yscale * self.figsize[1],
            )
        )

    # =============================================================================

    def colorbar(
        self,
        name,
        zrange=(0, 1),
        label="",
        log=False,
        tickcolor=None,
        orientation="right",
        pos=None,
        shrink=1.0,
    ):
        r"""Places a colorbar adjacent to the current figure.

        Parameters
        ----------
        name : string
            name of the (matplotlib) colormap to use
        zrange : tuple of floats
            min and max of the colorbar's range
        label : string
            colorbar label
        log : boolean
            Flag to use a logarithmic scale
        tickcolor : `pyx.color.*.*`
            Color for the tickmarks.  Example: pyx.color.cmyk.black
        orientation : string
            Placement of the colorbar.  Can be "left", "right", "top",
            or "bottom".
        pos : list of floats
            (x,y) position of the origin of the colorbar in centimeters.
        shrink : float
            Factor to shrink the colorbar's size.  A value of 1 means the
            colorbar will have a height / width of the figure.

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("image.jpg")
        >>> d.colorbar(
        ...     "hot", xrange=(1e-2, 1e-4), log=True, label="Density [cm$^{-3}$]"
        ... )
        >>> d.save_fig()
        """
        if pos is None:
            pos = [0, 0]

        if orientation == "right":
            origin = (pos[0] + self.figsize[0] + 0.5, pos[1])
            size = (0.1 * self.figsize[0], self.figsize[1])
            imsize = (1, 256)
        elif orientation == "left":
            origin = (pos[0] - 0.5 - 0.1 * self.figsize[0], pos[1])
            size = (0.1 * self.figsize[0], self.figsize[1])
            imsize = (1, 256)
        elif orientation == "top":
            origin = (pos[0], pos[1] + self.figsize[1] + 0.5)
            imorigin = (pos[0] + self.figsize[0], pos[1] + self.figsize[1] + 0.5)
            size = (self.figsize[0], 0.1 * self.figsize[1])
            imsize = (256, 1)
        elif orientation == "bottom":
            origin = (pos[0], pos[1] - 0.5 - 0.1 * self.figsize[1])
            imorigin = (pos[0] + self.figsize[0], pos[1] - 0.5 - 0.1 * self.figsize[1])
            size = (self.figsize[0], 0.1 * self.figsize[1])
            imsize = (256, 1)
        else:
            raise RuntimeError(f"orientation {orientation} unknown")

        # If shrink is a scalar, then convert into tuple
        if not isinstance(shrink, (tuple, list)):
            shrink = (shrink, shrink)

        # Scale the colorbar
        shift = (0.5 * (1.0 - shrink[0]) * size[0], 0.5 * (1.0 - shrink[1]) * size[1])
        # To facilitate stretching rather than shrinking
        # If stretched in both directions (makes no sense?) then y dominates.
        if shrink[0] > 1.0:
            shift = (0.05 * self.figsize[0], 0.5 * (1.0 - shrink[1]) * size[1])
        if shrink[1] > 1.0:
            shift = (0.5 * (1.0 - shrink[0]) * size[0], 0.05 * self.figsize[1])
        size = (size[0] * shrink[0], size[1] * shrink[1])
        origin = (origin[0] + shift[0], origin[1] + shift[1])

        # Convert the colormap into a string
        x = np.linspace(1, 0, 256)
        cm_string = cm.get_cmap[name](x, bytes=True)[:, 0:3].tobytes()

        cmap_im = pyx.bitmap.image(imsize[0], imsize[1], "RGB", cm_string)
        if orientation == "top" or orientation == "bottom":
            imorigin = (imorigin[0] - shift[0], imorigin[1] + shift[1])
            self.canvas.insert(
                pyx.bitmap.bitmap(
                    imorigin[0], imorigin[1], cmap_im, width=-size[0], height=size[1]
                )
            )
        else:
            self.canvas.insert(
                pyx.bitmap.bitmap(
                    origin[0], origin[1], cmap_im, width=size[0], height=size[1]
                )
            )

        if tickcolor is None:
            c1 = pyx.graph.axis.painter.regular(tickattrs=[pyx.color.cmyk.black])
            pyx.graph.axis.painter.regular(
                tickattrs=[pyx.color.cmyk.black], labelattrs=None
            )
        else:
            c1 = pyx.graph.axis.painter.regular(tickattrs=[tickcolor])
            pyx.graph.axis.painter.regular(tickattrs=[tickcolor], labelattrs=None)
        if log:
            yaxis = pyx.graph.axis.log(
                min=zrange[0], max=zrange[1], title=label, painter=c1
            )
            yaxis2 = pyx.graph.axis.log(min=zrange[0], max=zrange[1], parter=None)
        else:
            yaxis = pyx.graph.axis.lin(
                min=zrange[0], max=zrange[1], title=label, painter=c1
            )
            yaxis2 = pyx.graph.axis.lin(min=zrange[0], max=zrange[1], parter=None)
        xaxis = pyx.graph.axis.lin(parter=None)

        if orientation == "right":
            _colorbar = pyx.graph.graphxy(
                width=size[0],
                height=size[1],
                xpos=origin[0],
                ypos=origin[1],
                x=xaxis,
                y=yaxis2,
                y2=yaxis,
            )
        elif orientation == "left":
            _colorbar = pyx.graph.graphxy(
                width=size[0],
                height=size[1],
                xpos=origin[0],
                ypos=origin[1],
                x=xaxis,
                y2=yaxis2,
                y=yaxis,
            )
        elif orientation == "top":
            _colorbar = pyx.graph.graphxy(
                width=size[0],
                height=size[1],
                xpos=origin[0],
                ypos=origin[1],
                y=xaxis,
                x=yaxis2,
                x2=yaxis,
            )
        elif orientation == "bottom":
            _colorbar = pyx.graph.graphxy(
                width=size[0],
                height=size[1],
                xpos=origin[0],
                ypos=origin[1],
                y=xaxis,
                x2=yaxis2,
                x=yaxis,
            )

        blank_data = pyx.graph.data.points([(-1e10, -1e10), (-9e10, -9e10)], x=1, y=2)
        _colorbar.plot(blank_data)
        self.canvas.insert(_colorbar)

    # =============================================================================

    def colorbar_yt(self, plot, field=None, cb_labels=None, **kwargs):
        r"""Wrapper around DualEPS.colorbar to take information from a yt plot.

        Accepts all parameters that DualEPS.colorbar takes.

        Parameters
        ----------
        plot : A yt plot
            yt plot from which the information is taken.
        cb_labels : list of labels for the colorbars. List should be the same
                    size as the number of colorbars used. Should be passed
                    into this function by either the singleplot or multiplot api.

        Examples
        --------
        >>> p = SlicePlot(ds, 0, "density")
        >>> p.hide_colorbar()
        >>> d = DualEPS()
        >>> d.axis_box_yt(p)
        >>> d.insert_image_yt(p)
        >>> d.colorbar_yt(p)
        >>> d.save_fig()
        """

        if isinstance(plot, ProfilePlot):
            raise RuntimeError(
                "When using ProfilePlots you must either set yt_nocbar=True or provide "
                "colorbar flags so that the profiles don't have colorbars"
            )
        _cmap = None
        if field is not None:
            self.field = plot.data_source._determine_fields(field)[0]
        if isinstance(plot, (PlotWindow, PhasePlot)):
            _cmap = plot._colormap_config[self.field]
        else:
            if plot.cmap is not None:
                _cmap = plot.cmap.name
        if _cmap is None:
            _cmap = ytcfg.get("yt", "default_colormap")
        if isinstance(plot, (PlotWindow, PhasePlot)):
            if isinstance(plot, PlotWindow):
                try:
                    _zlabel = plot.frb[self.field].info["label"]
                    _unit = Unit(
                        plot.frb[self.field].units, registry=plot.ds.unit_registry
                    )
                    units = _unit.latex_representation()
                    # PyX does not support \frac because it's based on TeX.
                    units = pyxize_label(units)
                    _zlabel += r" (" + units + r")"
                except NotImplementedError:
                    print("Colorbar label not available")
                    _zlabel = ""
            else:
                _, _, z_title = plot._get_field_title(self.field, plot.profile)
                _zlabel = pyxize_label(z_title)
            _zlabel = _zlabel.replace("_", r"\;")
            _zlog = plot.get_log(self.field)[self.field]
            if plot.plots[self.field].zmin is None:
                zmin = plot.plots[self.field].image._A.min()
            else:
                zmin = plot.plots[self.field].zmin
            if plot.plots[self.field].zmax is None:
                zmax = plot.plots[self.field].image._A.max()
            else:
                zmax = plot.plots[self.field].zmax
            _zrange = (zmin, zmax)
        else:
            _zlabel = plot._z_label.replace("_", r"\;")
            _zlog = plot._log_z
            _zrange = (plot.norm.vmin, plot.norm.vmax)
        if cb_labels is not None:  # Overrides deduced labels
            _zlabel = cb_labels.pop()
        self.colorbar(_cmap, zrange=_zrange, label=_zlabel, log=_zlog, **kwargs)

    # =============================================================================

    def circle(
        self,
        radius=0.2,
        loc=(0.5, 0.5),
        color=pyx.color.cmyk.white,
        linewidth=pyx.style.linewidth.normal,
    ):
        r"""Draws a circle in the current figure.

        Parameters
        ----------
        radius : float
            Radius of the circle in units of figsize
        loc : tuple of floats
            Location of the circle's center in units of figsize
        color : `pyx.color.*.*`
            Color of the circle stroke.  Example: pyx.color.cmyk.white
        linewidth : `pyx.style.linewidth.*`
            Width of the circle stroke width. Example:
            pyx.style.linewidth.normal

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("image.jpg")
        >>> d.circle(radius=0.1, color=pyx.color.cmyk.Red)
        >>> d.save_fig()
        """
        circle = pyx.path.circle(
            self.figsize[0] * loc[0], self.figsize[1] * loc[1], self.figsize[0] * radius
        )
        self.canvas.stroke(circle, [color, linewidth])

    # =============================================================================

    def arrow(
        self,
        size=0.2,
        label="",
        loc=(0.05, 0.08),
        labelloc="top",
        color=pyx.color.cmyk.white,
        linewidth=pyx.style.linewidth.normal,
        rotation=0.0,
    ):
        r"""Draws an arrow in the current figure

        Parameters
        ----------
        size : float
            Length of arrow (base to tip) in units of the figure size.
        label : string
            Annotation label of the arrow.
        loc : tuple of floats
            Location of the left hand side of the arrow in units of
            the figure size.
        rotation : float
            Orientation angle of the arrow in units of degrees
        labelloc : string
            Location of the label with respect to the line.  Can be
            "top" or "bottom"
        color : `pyx.color.*.*`
            Color of the arrow.  Example: pyx.color.cymk.white
        linewidth : `pyx.style.linewidth.*`
            Width of the arrow.  Example: pyx.style.linewidth.normal

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("arrow_image.jpg")
        >>> d.arrow(size=0.2, label="Black Hole!", loc=(0.05, 0.1))
        >>> d.save_fig()
        """
        line = pyx.path.line(
            self.figsize[0] * loc[0],
            self.figsize[1] * loc[1],
            self.figsize[0] * (loc[0] + size * np.cos(np.pi * rotation / 180)),
            self.figsize[1] * (loc[1] + size * np.sin(np.pi * rotation / 180)),
        )
        self.canvas.stroke(line, [linewidth, color, pyx.deco.earrow()])

        if labelloc == "bottom":
            yoff = -0.1 * size
            valign = pyx.text.valign.top
        else:
            yoff = +0.1 * size
            valign = pyx.text.valign.bottom
        if label != "":
            self.canvas.text(
                self.figsize[0] * (loc[0] + 0.5 * size),
                self.figsize[1] * (loc[1] + yoff),
                label,
                [color, valign, pyx.text.halign.center],
            )

    # =============================================================================

    def scale_line(
        self,
        size=0.2,
        label="",
        loc=(0.05, 0.08),
        labelloc="top",
        color=pyx.color.cmyk.white,
        linewidth=pyx.style.linewidth.normal,
    ):
        r"""Draws a scale line in the current figure.

        Parameters
        ----------
        size : float
            Length of the scale line in units of the figure size.
        label : string
            Annotation label of the scale line.
        loc : tuple of floats
            Location of the left hand side of the scale line in units of
            the figure size.
        labelloc : string
            Location of the label with respect to the line.  Can be
            "top" or "bottom"
        color : `pyx.color.*.*`
            Color of the scale line.  Example: pyx.color.cymk.white
        linewidth : `pyx.style.linewidth.*`
            Width of the scale line.  Example: pyx.style.linewidth.normal

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("image.jpg")
        >>> d.scale_line(size=0.2, label="1 kpc", loc=(0.05, 0.1))
        >>> d.save_fig()
        """

        line = pyx.path.line(
            self.figsize[0] * loc[0],
            self.figsize[1] * loc[1],
            self.figsize[0] * (loc[0] + size),
            self.figsize[1] * loc[1],
        )
        self.canvas.stroke(line, [linewidth, color])
        line = pyx.path.line(
            self.figsize[0] * loc[0],
            self.figsize[1] * (loc[1] - 0.1 * size),
            self.figsize[0] * loc[0],
            self.figsize[1] * (loc[1] + 0.1 * size),
        )
        self.canvas.stroke(line, [linewidth, color])
        line = pyx.path.line(
            self.figsize[0] * (loc[0] + size),
            self.figsize[1] * (loc[1] - 0.1 * size),
            self.figsize[0] * (loc[0] + size),
            self.figsize[1] * (loc[1] + 0.1 * size),
        )
        self.canvas.stroke(line, [linewidth, color])

        if labelloc == "bottom":
            yoff = -0.1 * size
            valign = pyx.text.valign.top
        else:
            yoff = +0.1 * size
            valign = pyx.text.valign.bottom
        if label != "":
            self.canvas.text(
                self.figsize[0] * (loc[0] + 0.5 * size),
                self.figsize[1] * (loc[1] + yoff),
                label,
                [color, valign, pyx.text.halign.center],
            )

    # =============================================================================

    def title_box(
        self,
        text,
        color=pyx.color.cmyk.black,
        bgcolor=pyx.color.cmyk.white,
        loc=(0.02, 0.98),
        halign=pyx.text.halign.left,
        valign=pyx.text.valign.top,
        text_opts=None,
    ):
        r"""Inserts a box with text in the current figure.

        Parameters
        ----------
        text : string
            String to insert in the textbox.
        color : `pyx.color.*.*`
            Color of the text.  Example: pyx.color.cmyk.black
        bgcolor : `pyx.color.*.*`
            Color of the textbox background.  Example: pyx.color.cmyk.white
        loc : tuple of floats
            Location of the textbox origin in units of the figure size.
        halign : `pyx.text.halign.*`
            Horizontal alignment of the text.  Example: pyx.text.halign.left
        valign : `pyx.text.valign.*`
            Vertical alignment of the text.  Example: pyx.text.valign.top

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        >>> d.insert_image("image.jpg")
        >>> d.title_box("Halo 1", loc=(0.05, 0.95))
        >>> d.save_fig()
        """
        if text_opts is None:
            text_opts = []
        tbox = self.canvas.text(
            self.figsize[0] * loc[0],
            self.figsize[1] * loc[1],
            text,
            [color, valign, halign] + text_opts,
        )
        if bgcolor is not None:
            tpath = tbox.bbox().enlarged(2 * pyx.unit.x_pt).path()
            self.canvas.draw(tpath, [pyx.deco.filled([bgcolor]), pyx.deco.stroked()])
        self.canvas.insert(tbox)

    # =============================================================================

    def save_fig(self, filename="test", format="eps", resolution=250):
        r"""Saves current figure to a file.

        Parameters
        ----------
        filename : string
            Name of the saved file without the extension.
        format : string
            Format type.  Can be "eps" or "pdf"

        Examples
        --------
        >>> d = DualEPS()
        >>> d.axis_box(xrange=(0, 100), yrange=(1e-3, 1), ylog=True)
        """
        if format == "eps":
            self.canvas.writeEPSfile(filename)
        elif format == "pdf":
            self.canvas.writePDFfile(filename)
        elif format == "png":
            self.canvas.writeGSfile(filename + ".png", "png16m", resolution=resolution)
        elif format == "jpg":
            self.canvas.writeGSfile(filename + ".jpeg", "jpeg", resolution=resolution)
        else:
            raise RuntimeError(f"format {format} unknown.")


# =============================================================================
# =============================================================================
# =============================================================================


def multiplot(
    ncol,
    nrow,
    yt_plots=None,
    fields=None,
    images=None,
    xranges=None,
    yranges=None,
    xlabels=None,
    ylabels=None,
    xdata=None,
    ydata=None,
    colorbars=None,
    shrink_cb=0.95,
    figsize=(8, 8),
    margins=(0, 0),
    titles=None,
    savefig=None,
    format="eps",
    yt_nocbar=False,
    bare_axes=False,
    xaxis_flags=None,
    yaxis_flags=None,
    cb_flags=None,
    cb_location=None,
    cb_labels=None,
):
    r"""Convenience routine to create a multi-panel figure from yt plots or
    JPEGs.  The images are first placed from the origin, and then
    bottom-to-top and left-to-right.

    Parameters
    ----------
    ncol : integer
        Number of columns in the figure.
    nrow : integer
        Number of rows in the figure.
    yt_plots : list of yt plot instances
        yt plots to include in the figure.
    images : list of strings
        JPEG filenames to include in the figure.
    xranges : list of tuples
        The min and max of the x-axes
    yranges : list of tuples
        The min and max of the y-axes
    xlabels : list of strings
        Labels for the x-axes
    ylabels : list of strings
        Labels for the y-axes
    colorbars : list of dicts
        Dicts that describe the type of colorbar to be used in each
        figure.  Use the function return_colormap() to create these dicts.
    shrink_cb : float
        Factor by which the colorbar is shrunk.
    figsize : tuple of floats
        The width and height of a single figure in centimeters.
    margins : tuple of floats
        The horizontal and vertical margins between panels in centimeters.
    titles : list of strings
        Titles that are placed in textboxes in each panel.
    savefig : string
        Name of the saved file without the extension.
    format : string
        File format of the figure. eps or pdf accepted.
    yt_nocbar : boolean
        Flag to indicate whether or not colorbars are created.
    bare_axes : boolean
        Set to true to have no annotations or tick marks on all of the
        axes.
    cb_flags : list of booleans
        Flags for each plot to have a colorbar or not.
    cb_location : list of strings
        Strings to control the location of the colorbar (left, right,
        top, bottom)
    cb_labels : list of labels for the colorbars. List should be the same
                size as the number of colorbars used.

    Examples
    --------
    >>> images = ["density.jpg", "hi_density.jpg", "entropy.jpg", "special.jpg"]
    >>> cbs = []
    >>> cbs.append(return_colormap("arbre", "Density [cm$^{-3}$]", (0, 10), False))
    >>> cbs.append(return_colormap("kelp", "HI Density", (0, 5), False))
    >>> cbs.append(return_colormap("hot", r"Entropy [K cm$^2$]", (1e-2, 1e6), True))
    >>> cbs.append(return_colormap("Spectral", "Stuff$_x$!", (1, 300), True))
    >>> mp = multiplot(
    ...     2,
    ...     2,
    ...     images=images,
    ...     margins=(0.1, 0.1),
    ...     titles=["1", "2", "3", "4"],
    ...     xlabels=["one", "two"],
    ...     ylabels=None,
    ...     colorbars=cbs,
    ...     shrink_cb=0.95,
    ... )
    >>> mp.scale_line(label="$r_{vir}$", labelloc="top")
    >>> mp.save_fig("multiplot")

    Notes
    -----
    If given both yt_plots and images, this will get preference to the
    yt plots.
    """
    # Error check
    npanels = ncol * nrow
    if cb_labels is not None:
        cb_labels.reverse()  # Because I pop the list

    if images is not None:
        if len(images) != npanels:
            raise RuntimeError(
                "Number of images (%d) doesn't match nrow(%d)"
                " x ncol(%d)." % (len(images), nrow, ncol)
            )
    if yt_plots is None and images is None:
        raise RuntimeError("Must supply either yt_plots or image filenames.")
    if yt_plots is not None and images is not None:
        mylog.warning("Given both images and yt plots.  Ignoring images.")
    if yt_plots is not None:
        _yt = True
    else:
        _yt = False
    if fields is None:
        fields = [None] * npanels

    # If no ranges or labels given and given only images, fill them in.
    if not _yt:
        if xranges is None:
            xranges = []
            for _ in range(npanels):
                xranges.append((0, 1))
        if yranges is None:
            yranges = []
            for _ in range(npanels):
                yranges.append((0, 1))
        if xlabels is None:
            xlabels = []
            for _ in range(npanels):
                xlabels.append("")
        if ylabels is None:
            ylabels = []
            for _ in range(npanels):
                ylabels.append("")

    d = DualEPS(figsize=figsize)
    for j in range(nrow):
        invj = nrow - j - 1
        ypos = invj * (figsize[1] + margins[1])
        for i in range(ncol):
            xpos = i * (figsize[0] + margins[0])
            index = j * ncol + i
            if isinstance(yt_plots, list):
                this_plot = yt_plots[index]
            else:
                this_plot = yt_plots
            if j == nrow - 1:
                xaxis = 0
            elif j == 0:
                xaxis = 1
            else:
                xaxis = -1
            if i == 0:
                yaxis = 0
            elif i == ncol - 1:
                yaxis = 1
            else:
                yaxis = -1
            if xdata is None:
                _xdata = None
            else:
                _xdata = xdata[index]
            if ydata is None:
                _ydata = None
            else:
                _ydata = ydata[index]
            if xaxis_flags is not None:
                if xaxis_flags[index] is not None:
                    xaxis = xaxis_flags[index]
            if yaxis_flags is not None:
                if yaxis_flags[index] is not None:
                    yaxis = yaxis_flags[index]
            if _yt:
                this_plot._setup_plots()
                if xlabels is not None:
                    xlabel = xlabels[i]
                else:
                    xlabel = None
                if ylabels is not None:
                    ylabel = ylabels[j]
                else:
                    ylabel = None
                d.insert_image_yt(this_plot, pos=(xpos, ypos), field=fields[index])
                d.axis_box_yt(
                    this_plot,
                    pos=(xpos, ypos),
                    bare_axes=bare_axes,
                    xaxis_side=xaxis,
                    yaxis_side=yaxis,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xdata=_xdata,
                    ydata=_ydata,
                )
            else:
                d.insert_image(images[index], pos=(xpos, ypos))
                d.axis_box(
                    pos=(xpos, ypos),
                    xrange=xranges[index],
                    yrange=yranges[index],
                    xlabel=xlabels[i],
                    ylabel=ylabels[j],
                    bare_axes=bare_axes,
                    xaxis_side=xaxis,
                    yaxis_side=yaxis,
                    xdata=_xdata,
                    ydata=_ydata,
                )
            if titles is not None:
                if titles[index] is not None:
                    d.title_box(
                        titles[index],
                        loc=(
                            i + 0.05 + i * margins[0] / figsize[0],
                            j + 0.98 + j * margins[1] / figsize[1],
                        ),
                    )

    # Insert colorbars after all axes are placed because we want to
    # put them on the edges of the bounding box.
    bbox = (
        100.0 * d.canvas.bbox().left().t,
        100.0 * d.canvas.bbox().right().t - d.figsize[0],
        100.0 * d.canvas.bbox().bottom().t,
        100.0 * d.canvas.bbox().top().t - d.figsize[1],
    )
    for j in range(nrow):
        invj = nrow - j - 1
        ypos0 = invj * (figsize[1] + margins[1])
        for i in range(ncol):
            xpos0 = i * (figsize[0] + margins[0])
            index = j * ncol + i
            if isinstance(yt_plots, list):
                this_plot = yt_plots[index]
            else:
                this_plot = yt_plots
            if (not _yt and colorbars is not None) or (_yt and not yt_nocbar):
                if cb_flags is not None:
                    if not cb_flags[index]:
                        continue
                if cb_location is None:
                    if ncol == 1:
                        orientation = "right"
                    elif i == 0:
                        orientation = "left"
                    elif i + 1 == ncol:
                        orientation = "right"
                    elif j == 0:
                        orientation = "bottom"
                    elif j + 1 == nrow:
                        orientation = "top"
                    else:
                        orientation = None  # Marker for interior plot
                else:
                    if isinstance(cb_location, dict):
                        if fields[index] not in cb_location.keys():
                            raise RuntimeError(
                                f"{fields[index]} not found in cb_location dict"
                            )
                        orientation = cb_location[fields[index]]
                    elif isinstance(cb_location, list):
                        orientation = cb_location[index]
                    else:
                        raise RuntimeError("Bad format: cb_location")
                if orientation == "right":
                    xpos = bbox[1]
                    ypos = ypos0
                elif orientation == "left":
                    xpos = bbox[0]
                    ypos = ypos0
                elif orientation == "bottom":
                    ypos = bbox[2]
                    xpos = xpos0
                elif orientation == "top":
                    ypos = bbox[3]
                    xpos = xpos0
                else:
                    mylog.warning(
                        "Unknown colorbar location %s. No colorbar displayed.",
                        orientation,
                    )
                    orientation = None  # Marker for interior plot

                if orientation is not None:
                    if _yt:
                        # Set field if undefined
                        if fields[index] is None:
                            fields[index] = d.return_field(yt_plots[index])

                        d.colorbar_yt(
                            this_plot,
                            field=fields[index],
                            pos=[xpos, ypos],
                            shrink=shrink_cb,
                            orientation=orientation,
                            cb_labels=cb_labels,
                        )
                    else:
                        d.colorbar(
                            colorbars[index]["cmap"],
                            zrange=colorbars[index]["range"],
                            label=colorbars[index]["name"],
                            log=colorbars[index]["log"],
                            orientation=orientation,
                            pos=[xpos, ypos],
                            shrink=shrink_cb,
                        )

    if savefig is not None:
        d.save_fig(savefig, format=format)

    return d


# =============================================================================


def multiplot_yt(ncol, nrow, plots, fields=None, **kwargs):
    r"""Wrapper for multiplot that takes a yt PlotWindow

    Accepts all parameters used in multiplot.

    Parameters
    ----------
    ncol : integer
        Number of columns in the figure.
    nrow : integer
        Number of rows in the figure.
    plots : ``PlotWindow`` instance, ``PhasePlot`` instance, or list of plots
        yt plots to be used.

    Examples
    --------
    >>> p1 = SlicePlot(ds, 0, "density")
    >>> p1.set_width(10, "kpc")

    >>> p2 = SlicePlot(ds, 0, "temperature")
    >>> p2.set_width(10, "kpc")
    >>> p2.set_colormap("temperature", "hot")

    >>> sph = ds.sphere(ds.domain_center, (10, "kpc"))
    >>> p3 = PhasePlot(
    ...     sph, "radius", "density", "temperature", weight_field="cell_mass"
    ... )

    >>> p4 = PhasePlot(sph, "radius", "density", "pressure", "cell_mass")

    >>> mp = multiplot_yt(
    ...     2,
    ...     2,
    ...     [p1, p2, p3, p4],
    ...     savefig="yt",
    ...     shrink_cb=0.9,
    ...     bare_axes=True,
    ...     yt_nocbar=False,
    ...     margins=(0.5, 0.5),
    ... )
    """
    # Determine whether the plots are organized in a PlotWindow, or list
    # of PlotWindows
    if isinstance(plots, (PlotWindow, PhasePlot)):
        if fields is None:
            fields = plots.fields
        if len(fields) < nrow * ncol:
            raise RuntimeError(
                "Number of plots ({}) is less "
                "than nrow({}) x ncol({}).".format(len(fields), nrow, ncol)
            )
        figure = multiplot(ncol, nrow, yt_plots=plots, fields=fields, **kwargs)
    elif isinstance(plots, list) and isinstance(plots[0], (PlotWindow, PhasePlot)):
        if len(plots) < nrow * ncol:
            raise RuntimeError(
                "Number of plots ({}) is less "
                "than nrow({}) x ncol({}).".format(len(fields), nrow, ncol)
            )
        figure = multiplot(ncol, nrow, yt_plots=plots, fields=fields, **kwargs)
    else:
        raise RuntimeError("Unknown plot type in multiplot_yt")
    return figure


# =============================================================================


def single_plot(
    plot,
    field=None,
    figsize=(12, 12),
    cb_orient="right",
    bare_axes=False,
    savefig=None,
    colorbar=True,
    file_format="eps",
    **kwargs,
):
    r"""Wrapper for DualEPS routines to create a figure directly from a yt
    plot.  Calls insert_image_yt, axis_box_yt, and colorbar_yt.

    Parameters
    ----------
    plot : `yt.visualization.plot_window.PlotWindow`
        yt plot that provides the image and metadata
    figsize : tuple of floats
        Size of the figure in centimeters.
    cb_orient : string
        Placement of the colorbar.  Can be "left", "right", "top", or
        "bottom".
    bare_axes : boolean
        Set to true to have no annotations or tick marks on all of the axes.
    savefig : string
        Name of the saved file without the extension.
    colorbar : boolean
        Set to true to include a colorbar
    file_format : string
        Format type.  Can be "eps" or "pdf"

    Examples
    --------
    >>> p = SlicePlot(ds, 0, "density")
    >>> p.set_width(0.1, "kpc")
    >>> single_plot(p, savefig="figure1")
    """
    d = DualEPS(figsize=figsize)
    d.insert_image_yt(plot, field=field)
    d.axis_box_yt(plot, bare_axes=bare_axes, **kwargs)
    if colorbar and not isinstance(plot, ProfilePlot):
        d.colorbar_yt(plot, orientation=cb_orient)
    if savefig is not None:
        d.save_fig(savefig, format=file_format)
    return d


# =============================================================================
def return_cmap(cmap=None, label="", range=(0, 1), log=False):
    issue_deprecation_warning(
        "Deprecated alias. Use return_colormap instead.", removal="4.1.0"
    )
    return return_colormap(cmap=cmap, label=label, crange=range, log=log)


def return_colormap(cmap=None, label="", range=(0, 1), log=False):
    r"""Returns a dict that describes a colorbar.  Exclusively for use with
    multiplot.

    Parameters
    ----------
    cmap : string
        name of the (matplotlib) colormap to use
    label : string
        colorbar label
    range : tuple of floats
        min and max of the colorbar's range
    log : boolean
        Flag to use a logarithmic scale

    Examples
    --------
    >>> cb = return_colormap("arbre", "Density [cm$^{-3}$]", (0, 10), False)
    """
    if cmap is None:
        cmap = ytcfg.get("yt", "default_colormap")
    return {"cmap": cmap, "name": label, "range": range, "log": log}



import numpy as np

from yt.utilities.lib.pixelization_routines import pixelize_cartesian, pixelize_cylinder

from .coordinate_handler import (
    CoordinateHandler,
    _get_coord_fields,
    _unknown_coord,
    cartesian_to_cylindrical,
    cylindrical_to_cartesian,
)

#
# Cylindrical fields
#


class CylindricalCoordinateHandler(CoordinateHandler):
    name = "cylindrical"

    def __init__(self, ds, ordering=("r", "z", "theta")):
        super().__init__(ds, ordering)
        self.image_units = {}
        self.image_units[self.axis_id["r"]] = ("rad", None)
        self.image_units[self.axis_id["theta"]] = (None, None)
        self.image_units[self.axis_id["z"]] = (None, None)

    def setup_fields(self, registry):
        # return the fields for r, z, theta
        registry.add_field(
            ("index", "dx"), sampling_type="cell", function=_unknown_coord
        )

        registry.add_field(
            ("index", "dy"), sampling_type="cell", function=_unknown_coord
        )

        registry.add_field(
            ("index", "x"), sampling_type="cell", function=_unknown_coord
        )

        registry.add_field(
            ("index", "y"), sampling_type="cell", function=_unknown_coord
        )
        f1, f2 = _get_coord_fields(self.axis_id["r"])
        registry.add_field(
            ("index", "dr"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="code_length",
        )

        registry.add_field(
            ("index", "r"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="code_length",
        )

        f1, f2 = _get_coord_fields(self.axis_id["z"])
        registry.add_field(
            ("index", "dz"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="code_length",
        )

        registry.add_field(
            ("index", "z"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="code_length",
        )

        f1, f2 = _get_coord_fields(self.axis_id["theta"], "")
        registry.add_field(
            ("index", "dtheta"),
            sampling_type="cell",
            function=f1,
            display_field=False,
            units="",
        )

        registry.add_field(
            ("index", "theta"),
            sampling_type="cell",
            function=f2,
            display_field=False,
            units="",
        )

        def _CylindricalVolume(field, data):
            r = data["index", "r"]
            dr = data["index", "dr"]
            vol = 0.5 * ((r + 0.5 * dr) ** 2 - (r - 0.5 * dr) ** 2)
            vol *= data["index", "dtheta"]
            vol *= data["index", "dz"]
            return vol

        registry.add_field(
            ("index", "cell_volume"),
            sampling_type="cell",
            function=_CylindricalVolume,
            units="code_length**3",
        )
        registry.alias(("index", "volume"), ("index", "cell_volume"))

        def _path_r(field, data):
            return data["index", "dr"]

        registry.add_field(
            ("index", "path_element_r"),
            sampling_type="cell",
            function=_path_r,
            units="code_length",
        )

        def _path_theta(field, data):
            # Note: this already assumes cell-centered
            return data["index", "r"] * data["index", "dtheta"]

        registry.add_field(
            ("index", "path_element_theta"),
            sampling_type="cell",
            function=_path_theta,
            units="code_length",
        )

        def _path_z(field, data):
            return data["index", "dz"]

        registry.add_field(
            ("index", "path_element_z"),
            sampling_type="cell",
            function=_path_z,
            units="code_length",
        )

    def pixelize(
        self,
        dimension,
        data_source,
        field,
        bounds,
        size,
        antialias=True,
        periodic=False,
    ):
        # Note that above, we set periodic by default to be *false*.  This is
        # because our pixelizers, at present, do not handle periodicity
        # correctly, and if you change the "width" of a cylindrical plot, it
        # double-counts in the edge buffers.  See, for instance, issue 1669.
        ax_name = self.axis_name[dimension]
        if ax_name in ("r", "theta"):
            return self._ortho_pixelize(
                data_source, field, bounds, size, antialias, dimension, periodic
            )
        elif ax_name == "z":
            return self._cyl_pixelize(data_source, field, bounds, size, antialias)
        else:
            # Pixelizing along a cylindrical surface is a bit tricky
            raise NotImplementedError

    def _ortho_pixelize(
        self, data_source, field, bounds, size, antialias, dim, periodic
    ):
        period = self.period[:2].copy()  # dummy here
        period[0] = self.period[self.x_axis[dim]]
        period[1] = self.period[self.y_axis[dim]]
        if hasattr(period, "in_units"):
            period = period.in_units("code_length").d
        buff = np.zeros(size, dtype="f8")
        pixelize_cartesian(
            buff,
            data_source["px"],
            data_source["py"],
            data_source["pdx"],
            data_source["pdy"],
            data_source[field],
            bounds,
            int(antialias),
            period,
            int(periodic),
        )
        return buff

    def _cyl_pixelize(self, data_source, field, bounds, size, antialias):
        buff = np.full((size[1], size[0]), np.nan, dtype="f8")
        pixelize_cylinder(
            buff,
            data_source["px"],
            data_source["pdx"],
            data_source["py"],
            data_source["pdy"],
            data_source[field],
            bounds,
        )
        self.sanitize_buffer_fill_values(buff)
        return buff

    _x_pairs = (("r", "theta"), ("z", "r"), ("theta", "r"))
    _y_pairs = (("r", "z"), ("z", "theta"), ("theta", "z"))

    _image_axis_name = None

    @property
    def image_axis_name(self):
        if self._image_axis_name is not None:
            return self._image_axis_name
        # This is the x and y axes labels that get displayed.  For
        # non-Cartesian coordinates, we usually want to override these for
        # Cartesian coordinates, since we transform them.
        rv = {
            self.axis_id["r"]: ("theta", "z"),
            self.axis_id["z"]: ("x", "y"),
            self.axis_id["theta"]: ("r", "z"),
        }
        for i in list(rv.keys()):
            rv[self.axis_name[i]] = rv[i]
            rv[self.axis_name[i].upper()] = rv[i]
        self._image_axis_name = rv
        return rv

    def convert_from_cartesian(self, coord):
        return cartesian_to_cylindrical(coord)

    def convert_to_cartesian(self, coord):
        return cylindrical_to_cartesian(coord)

    def convert_to_cylindrical(self, coord):
        return coord

    def convert_from_cylindrical(self, coord):
        return coord

    def convert_to_spherical(self, coord):
        raise NotImplementedError

    def convert_from_spherical(self, coord):
        raise NotImplementedError

    @property
    def period(self):
        return np.array([0.0, 0.0, 2.0 * np.pi])

    def sanitize_center(self, center, axis):
        center, display_center = super().sanitize_center(center, axis)
        display_center = [
            0.0 * display_center[0],
            0.0 * display_center[1],
            0.0 * display_center[2],
        ]
        ax_name = self.axis_name[axis]
        r_ax = self.axis_id["r"]
        theta_ax = self.axis_id["theta"]
        z_ax = self.axis_id["z"]
        if ax_name == "r":
            display_center[theta_ax] = self.ds.domain_center[theta_ax]
            display_center[z_ax] = self.ds.domain_center[z_ax]
        elif ax_name == "theta":
            # use existing center value
            for idx in (r_ax, z_ax):
                display_center[idx] = center[idx]
        return center, display_center

    def sanitize_width(self, axis, width, depth):
        name = self.axis_name[axis]
        r_ax, theta_ax, z_ax = (
            self.ds.coordinates.axis_id[ax] for ax in ("r", "theta", "z")
        )
        if width is not None:
            width = super().sanitize_width(axis, width, depth)
        # Note: regardless of axes, these are set up to give consistent plots
        # when plotted, which is not strictly a "right hand rule" for axes.
        elif name == "r":  # soup can label
            width = [2.0 * np.pi * self.ds.domain_width.uq, self.ds.domain_width[z_ax]]
        elif name == "theta":
            width = [self.ds.domain_width[r_ax], self.ds.domain_width[z_ax]]
        elif name == "z":
            width = [
                2.0 * self.ds.domain_right_edge[r_ax],
                2.0 * self.ds.domain_right_edge[r_ax],
            ]
        return width



import numpy as np
from nose.tools import assert_raises

from yt.testing import (
    assert_almost_equal,
    assert_equal,
    fake_amr_ds,
    fake_particle_ds,
    fake_random_ds,
    requires_file,
)
from yt.utilities.answer_testing.framework import data_dir_load
from yt.utilities.exceptions import YTDimensionalityError
from yt.visualization.line_plot import LineBuffer

# This will test the "dataset access" method.


def test_box_creation():
    ds = fake_random_ds(32, length_unit=2)
    left_edge = ds.arr([0.2, 0.2, 0.2], "cm")
    right_edge = ds.arr([0.6, 0.6, 0.6], "cm")
    center = (left_edge + right_edge) / 2

    boxes = [
        ds.box(left_edge, right_edge),
        ds.box(0.5 * np.array(left_edge), 0.5 * np.array(right_edge)),
        ds.box((0.5 * left_edge).tolist(), (0.5 * right_edge).tolist()),
    ]

    region = ds.region(center, left_edge, right_edge)

    for b in boxes:
        assert_almost_equal(b.left_edge, region.left_edge)
        assert_almost_equal(b.right_edge, region.right_edge)
        assert_almost_equal(b.center, region.center)


def test_region_from_d():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    # We'll do a couple here

    # First, no string units
    reg1 = ds.r[0.2:0.3, 0.4:0.6, :]
    reg2 = ds.region([0.25, 0.5, 0.5], [0.2, 0.4, 0.0], [0.3, 0.6, 1.0])
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, "cm"):(0.5, "cm"), :, (0.25, "cm"):(0.35, "cm")]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])

    # Now, string units in some -- 1.0 == cm
    reg1 = ds.r[(0.1, "cm"):(0.5, "cm"), :, 0.25:0.35]
    reg2 = ds.region([0.3, 0.5, 0.3], [0.1, 0.0, 0.25], [0.5, 1.0, 0.35])
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])

    # And, lots of : usage!
    reg1 = ds.r[:, :, :]
    reg2 = ds.all_data()
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])

    # Test slice as an index
    reg1 = ds.r[0.1:0.8]
    reg2 = ds.region([0.45, 0.45, 0.45], [0.1, 0.1, 0.1], [0.8, 0.8, 0.8])
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])

    # Test with bad boundary initialization
    with assert_raises(RuntimeError):
        ds.r[0.3:0.1, 0.4:0.6, :]

    # Test region by creating an arbitrary grid
    reg1 = ds.r[0.15:0.55:16j, 0.25:0.65:32j, 0.35:0.75:64j]
    left_edge = np.array([0.15, 0.25, 0.35])
    right_edge = np.array([0.55, 0.65, 0.75])
    dims = np.array([16.0, 32.0, 64.0])
    reg2 = ds.arbitrary_grid(left_edge, right_edge, dims)
    assert_equal(reg1[("gas", "density")], reg2[("gas", "density")])


def test_accessing_all_data():
    # This will test first that we can access all_data, and next that we can
    # access it multiple times and get the *same object*.
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    dd = ds.all_data()
    assert_equal(ds.r[("gas", "density")], dd[("gas", "density")])
    # Now let's assert that it's the same object
    rho = ds.r[("gas", "density")]
    rho *= 2.0
    assert_equal(dd[("gas", "density")] * 2.0, ds.r[("gas", "density")])
    assert_equal(dd["gas", "density"] * 2.0, ds.r["gas", "density"])


def test_slice_from_r():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    sl1 = ds.r[0.5, :, :]
    sl2 = ds.slice("x", 0.5)
    assert_equal(sl1[("gas", "density")], sl2[("gas", "density")])

    frb1 = sl1.to_frb(width=1.0, height=1.0, resolution=(1024, 512))
    frb2 = ds.r[0.5, ::1024j, ::512j]
    assert_equal(frb1[("gas", "density")], frb2[("gas", "density")])

    # Test slice which doesn't cover the whole domain
    box = ds.box([0.0, 0.25, 0.25], [1.0, 0.75, 0.75])

    sl3 = ds.r[0.5, 0.25:0.75, 0.25:0.75]
    sl4 = ds.slice("x", 0.5, data_source=box)
    assert_equal(sl3[("gas", "density")], sl4[("gas", "density")])

    frb3 = sl3.to_frb(width=0.5, height=0.5, resolution=(1024, 512))
    frb4 = ds.r[0.5, 0.25:0.75:1024j, 0.25:0.75:512j]
    assert_equal(frb3[("gas", "density")], frb4[("gas", "density")])


def test_point_from_r():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    pt1 = ds.r[0.5, 0.3, 0.1]
    pt2 = ds.point([0.5, 0.3, 0.1])
    assert_equal(pt1[("gas", "density")], pt2[("gas", "density")])

    # Test YTDimensionalityError
    with assert_raises(YTDimensionalityError) as ex:
        ds.r[0.5, 0.1]
    assert_equal(str(ex.exception), "Dimensionality specified was 2 but we need 3")


def test_ray_from_r():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    ray1 = ds.r[(0.1, 0.2, 0.3):(0.4, 0.5, 0.6)]
    ray2 = ds.ray((0.1, 0.2, 0.3), (0.4, 0.5, 0.6))
    assert_equal(ray1[("gas", "density")], ray2[("gas", "density")])

    ray3 = ds.r[0.5 * ds.domain_left_edge : 0.5 * ds.domain_right_edge]
    ray4 = ds.ray(0.5 * ds.domain_left_edge, 0.5 * ds.domain_right_edge)
    assert_equal(ray3[("gas", "density")], ray4[("gas", "density")])

    start = [(0.1, "cm"), 0.2, (0.3, "cm")]
    end = [(0.5, "cm"), (0.4, "cm"), 0.6]
    ray5 = ds.r[start:end]
    start_arr = [ds.quan(0.1, "cm"), ds.quan(0.2, "cm"), ds.quan(0.3, "cm")]
    end_arr = [ds.quan(0.5, "cm"), ds.quan(0.4, "cm"), ds.quan(0.6, "cm")]
    ray6 = ds.ray(start_arr, end_arr)
    assert_equal(ray5[("gas", "density")], ray6[("gas", "density")])

    ray7 = ds.r[start:end:500j]
    ray8 = LineBuffer(ds, [0.1, 0.2, 0.3], [0.5, 0.4, 0.6], 500)
    assert_equal(ray7[("gas", "density")], ray8[("gas", "density")])


def test_ortho_ray_from_r():
    ds = fake_amr_ds(fields=["density"], units=["g/cm**3"])
    ray1 = ds.r[:, 0.3, 0.2]
    ray2 = ds.ortho_ray("x", [0.3, 0.2])
    assert_equal(ray1[("gas", "density")], ray2[("gas", "density")])

    # the y-coord is funny so test it too
    ray3 = ds.r[0.3, :, 0.2]
    ray4 = ds.ortho_ray("y", [0.2, 0.3])
    assert_equal(ray3[("gas", "density")], ray4[("gas", "density")])

    # Test ray which doesn't cover the whole domain
    box = ds.box([0.25, 0.0, 0.0], [0.75, 1.0, 1.0])
    ray5 = ds.r[0.25:0.75, 0.3, 0.2]
    ray6 = ds.ortho_ray("x", [0.3, 0.2], data_source=box)
    assert_equal(ray5[("gas", "density")], ray6[("gas", "density")])

    # Test fixed-resolution rays
    ray7 = ds.r[0.25:0.75:100j, 0.3, 0.2]
    ray8 = LineBuffer(ds, [0.2525, 0.3, 0.2], [0.7475, 0.3, 0.2], 100)
    assert_equal(ray7[("gas", "density")], ray8[("gas", "density")])


def test_particle_counts():
    ds = fake_random_ds(16, particles=100)
    assert ds.particle_type_counts == {"io": 100}

    pds = fake_particle_ds(npart=128)
    assert pds.particle_type_counts == {"io": 128}


g30 = "IsolatedGalaxy/galaxy0030/galaxy0030"


@requires_file(g30)
def test_checksum():
    assert fake_random_ds(16).checksum == "notafile"
    assert data_dir_load(g30).checksum == "6169536e4b9f737ce3d3ad440df44c58"



from collections import OrderedDict

import yt
from yt.frontends.gizmo.api import GizmoDataset
from yt.frontends.gizmo.fields import metal_elements
from yt.testing import requires_file
from yt.utilities.answer_testing.framework import requires_ds, sph_answer

# This maps from field names to weight field names to use for projections
fields = OrderedDict(
    [
        (("gas", "density"), None),
        (("gas", "temperature"), ("gas", "density")),
        (("gas", "metallicity"), ("gas", "density")),
        (("gas", "O_metallicity"), ("gas", "density")),
        (("gas", "velocity_magnitude"), None),
    ]
)

g64 = "gizmo_64/output/snap_N64L16_135.hdf5"
gmhd = "gizmo_mhd_mwdisk/gizmo_mhd_mwdisk.hdf5"
gmhd_bbox = [[-400, 400]] * 3


@requires_ds(g64, big_data=True)
def test_gizmo_64():
    ds = yt.load(g64)
    assert isinstance(ds, GizmoDataset)
    for test in sph_answer(ds, "snap_N64L16_135", 524288, fields):
        test_gizmo_64.__name__ = test.description
        yield test


@requires_file(gmhd)
def test_gizmo_mhd():
    """
    Magnetic fields should be loaded correctly when they are present.
    """
    ds = yt.load(gmhd, bounding_box=gmhd_bbox, unit_system="code")
    ad = ds.all_data()
    ptype = "PartType0"

    # Test vector magnetic field
    fmag = "particle_magnetic_field"
    f = ad[ptype, fmag]
    assert str(f.units) == "code_magnetic"
    assert f.shape == (409013, 3)

    # Test component magnetic fields
    for axis in "xyz":
        f = ad[ptype, fmag + "_" + axis]
        assert str(f.units) == "code_magnetic"
        assert f.shape == (409013,)


@requires_file(gmhd)
def test_gas_particle_fields():
    """
    Test fields set up in GizmoFieldInfo.setup_gas_particle_fields.
    """
    ds = yt.load(gmhd, bounding_box=gmhd_bbox)

    ptype = "PartType0"
    derived_fields = []
    # Add species fields
    for species in ["H_p0", "H_p1"]:
        for suffix in ["density", "fraction", "mass", "number_density"]:
            derived_fields += [f"{species}_{suffix}"]
    for species in metal_elements:
        derived_fields += [f"{species}_nuclei_mass_density"]
    # Add magnetic fields
    derived_fields += [f"particle_magnetic_field_{axis}" for axis in "xyz"]
    # Check
    for field in derived_fields:
        assert (ptype, field) in ds.derived_field_list

    ptype = "gas"
    derived_fields = []
    for species in ["H_p0", "H_p1"]:
        for suffix in ["density", "number_density"]:
            derived_fields += [f"{species}_{suffix}"]
    for species in metal_elements:
        for suffix in ["nuclei_mass_density", "metallicity"]:
            derived_fields += [f"{species}_{suffix}"]
    derived_fields += [f"magnetic_field_{axis}" for axis in "xyz"]
    for field in derived_fields:
        assert (ptype, field) in ds.derived_field_list


@requires_file(gmhd)
def test_star_particle_fields():
    """
    Test fields set up in GizmoFieldInfo.setup_star_particle_fields.
    """
    ds = yt.load(gmhd, bounding_box=gmhd_bbox)

    ptype = "PartType4"
    derived_fields = ["creation_time", "age"]
    for field in derived_fields:
        assert (ptype, field) in ds.derived_field_list



# These functions are RAMSES-specific
import re

from yt.config import ytcfg
from yt.funcs import mylog


def ramses_header(hvals):
    header = (
        ("ncpu", 1, "i"),
        ("ndim", 1, "i"),
        ("nx", 3, "i"),
        ("nlevelmax", 1, "i"),
        ("ngridmax", 1, "i"),
        ("nboundary", 1, "i"),
        ("ngrid_current", 1, "i"),
        ("boxlen", 1, "d"),
        ("nout", 3, "i"),
    )
    yield header
    # TODO: REMOVE
    noutput, iout, ifout = hvals["nout"]
    next_set = (
        ("tout", noutput, "d"),
        ("aout", noutput, "d"),
        ("t", 1, "d"),
        ("dtold", hvals["nlevelmax"], "d"),
        ("dtnew", hvals["nlevelmax"], "d"),
        ("nstep", 2, "i"),
        ("stat", 3, "d"),
        ("cosm", 7, "d"),
        ("timing", 5, "d"),
        ("mass_sph", 1, "d", True),
    )
    yield next_set


field_aliases = {
    "standard_five": ("Density", "x-velocity", "y-velocity", "z-velocity", "Pressure"),
    "standard_six": (
        "Density",
        "x-velocity",
        "y-velocity",
        "z-velocity",
        "Pressure",
        "Metallicity",
    ),
}

## Regular expressions used to parse file descriptors
VERSION_RE = re.compile(r"# version: *(\d+)")
# This will match comma-separated strings, discarding whitespaces
# on the left hand side
VAR_DESC_RE = re.compile(r"\s*([^\s]+),\s*([^\s]+),\s*([^\s]+)")

OUTPUT_DIR_RE = re.compile(r"(output|group)_(\d{5})")
STANDARD_FILE_RE = re.compile(r"((amr|hydro|part|grav)_\d{5}\.out\d{5}|info_\d{5}.txt)")


## Configure family mapping
particle_families = {
    "DM": 1,
    "star": 2,
    "cloud": 3,
    "dust": 4,
    "star_tracer": -2,
    "cloud_tracer": -3,
    "dust_tracer": -4,
    "gas_tracer": 0,
}

if ytcfg.has_section("ramses-families"):
    for key in particle_families.keys():
        val = ytcfg.get("ramses-families", key, fallback=None)
        if val is not None:
            mylog.info(
                "Changing family %s from %s to %s", key, particle_families[key], val
            )
            particle_families[key] = val



import numpy as np

from yt.fields.derived_field import ValidateParameter, ValidateSpatial
from yt.units.yt_array import uconcatenate, ucross
from yt.utilities.lib.misc_utilities import (
    obtain_position_vector,
    obtain_relative_velocity_vector,
)
from yt.utilities.math_utils import (
    get_cyl_r,
    get_cyl_r_component,
    get_cyl_theta,
    get_cyl_theta_component,
    get_cyl_z,
    get_cyl_z_component,
    get_sph_phi,
    get_sph_phi_component,
    get_sph_r_component,
    get_sph_theta,
    get_sph_theta_component,
    modify_reference_frame,
)

from .field_functions import get_radius
from .vector_operations import create_magnitude_field

sph_whitelist_fields = (
    "density",
    "temperature",
    "metallicity",
    "thermal_energy",
    "smoothing_length",
    "H_fraction",
    "He_fraction",
    "C_fraction",
    "Ca_fraction",
    "N_fraction",
    "O_fraction",
    "S_fraction",
    "Ne_fraction",
    "Mg_fraction",
    "Si_fraction",
    "Fe_fraction",
    "H_density",
    "He_density",
    "C_density",
    "Ca_density",
    "N_density",
    "O_density",
    "S_density",
    "Ne_density",
    "Mg_density",
    "Si_density",
    "Fe_density",
)


def _field_concat(fname):
    def _AllFields(field, data):
        v = []
        for ptype in data.ds.particle_types:
            data.ds._last_freq = (ptype, None)
            if ptype == "all" or ptype in data.ds.known_filters:
                continue
            v.append(data[ptype, fname].copy())
        rv = uconcatenate(v, axis=0)
        return rv

    return _AllFields


def _field_concat_slice(fname, axi):
    def _AllFields(field, data):
        v = []
        for ptype in data.ds.particle_types:
            data.ds._last_freq = (ptype, None)
            if ptype == "all" or ptype in data.ds.known_filters:
                continue
            v.append(data[ptype, fname][:, axi])
        rv = uconcatenate(v, axis=0)
        return rv

    return _AllFields


def particle_deposition_functions(ptype, coord_name, mass_name, registry):
    unit_system = registry.ds.unit_system
    orig = set(registry.keys())
    ptype_dn = ptype.replace("_", " ").title()

    def particle_count(field, data):
        pos = data[ptype, coord_name]
        d = data.deposit(pos, method="count")
        return data.apply_units(d, field.units)

    registry.add_field(
        ("deposit", f"{ptype}_count"),
        sampling_type="cell",
        function=particle_count,
        validators=[ValidateSpatial()],
        units="",
        display_name=r"\mathrm{%s Count}" % ptype_dn,
    )

    def particle_mass(field, data):
        pos = data[ptype, coord_name]
        pmass = data[ptype, mass_name]
        pmass.convert_to_units(field.units)
        d = data.deposit(pos, [pmass], method="sum")
        return data.apply_units(d, field.units)

    registry.add_field(
        ("deposit", f"{ptype}_mass"),
        sampling_type="cell",
        function=particle_mass,
        validators=[ValidateSpatial()],
        display_name=r"\mathrm{%s Mass}" % ptype_dn,
        units=unit_system["mass"],
    )

    def particle_density(field, data):
        pos = data[ptype, coord_name]
        pos.convert_to_units("code_length")
        mass = data[ptype, mass_name]
        mass.convert_to_units("code_mass")
        d = data.deposit(pos, [mass], method="sum")
        d = data.ds.arr(d, "code_mass")
        d /= data["index", "cell_volume"]
        return d

    registry.add_field(
        ("deposit", f"{ptype}_density"),
        sampling_type="cell",
        function=particle_density,
        validators=[ValidateSpatial()],
        display_name=r"\mathrm{%s Density}" % ptype_dn,
        units=unit_system["density"],
    )

    def particle_cic(field, data):
        pos = data[ptype, coord_name]
        d = data.deposit(pos, [data[ptype, mass_name]], method="cic")
        d = data.apply_units(d, data[ptype, mass_name].units)
        d /= data["index", "cell_volume"]
        return d

    registry.add_field(
        ("deposit", f"{ptype}_cic"),
        sampling_type="cell",
        function=particle_cic,
        validators=[ValidateSpatial()],
        display_name=r"\mathrm{%s CIC Density}" % ptype_dn,
        units=unit_system["density"],
    )

    def _get_density_weighted_deposit_field(fname, units, method):
        def _deposit_field(field, data):
            """
            Create a grid field for particle quantities weighted by particle
            mass, using cloud-in-cell deposit.
            """
            pos = data[ptype, "particle_position"]
            # Get back into density
            pden = data[ptype, "particle_mass"]
            top = data.deposit(pos, [pden * data[(ptype, fname)]], method=method)
            bottom = data.deposit(pos, [pden], method=method)
            top[bottom == 0] = 0.0
            bnz = bottom.nonzero()
            top[bnz] /= bottom[bnz]
            d = data.ds.arr(top, units=units)
            return d

        return _deposit_field

    for ax in "xyz":
        for method, name in zip(("cic", "sum"), ("cic", "nn")):
            function = _get_density_weighted_deposit_field(
                f"particle_velocity_{ax}", "code_velocity", method
            )
            registry.add_field(
                ("deposit", ("%s_" + name + "_velocity_%s") % (ptype, ax)),
                sampling_type="cell",
                function=function,
                units=unit_system["velocity"],
                take_log=False,
                validators=[ValidateSpatial(0)],
            )

    for method, name in zip(("cic", "sum"), ("cic", "nn")):
        function = _get_density_weighted_deposit_field("age", "code_time", method)
        registry.add_field(
            ("deposit", ("%s_" + name + "_age") % (ptype)),
            sampling_type="cell",
            function=function,
            units=unit_system["time"],
            take_log=False,
            validators=[ValidateSpatial(0)],
        )

    # Now some translation functions.

    def particle_ones(field, data):
        v = np.ones(data[ptype, coord_name].shape[0], dtype="float64")
        return data.apply_units(v, field.units)

    registry.add_field(
        (ptype, "particle_ones"),
        sampling_type="particle",
        function=particle_ones,
        units="",
        display_name=r"Particle Count",
    )

    def particle_mesh_ids(field, data):
        pos = data[ptype, coord_name]
        ids = np.zeros(pos.shape[0], dtype="float64") - 1
        # This is float64 in name only.  It will be properly cast inside the
        # deposit operation.
        # _ids = ids.view("float64")
        data.deposit(pos, [ids], method="mesh_id")
        return data.apply_units(ids, "")

    registry.add_field(
        (ptype, "mesh_id"),
        sampling_type="particle",
        function=particle_mesh_ids,
        validators=[ValidateSpatial()],
        units="",
    )

    return list(set(registry.keys()).difference(orig))


def particle_scalar_functions(ptype, coord_name, vel_name, registry):

    # Now we have to set up the various velocity and coordinate things.  In the
    # future, we'll actually invert this and use the 3-component items
    # elsewhere, and stop using these.

    # Note that we pass in _ptype here so that it's defined inside the closure.

    def _get_coord_funcs(axi, _ptype):
        def _particle_velocity(field, data):
            return data[_ptype, vel_name][:, axi]

        def _particle_position(field, data):
            return data[_ptype, coord_name][:, axi]

        return _particle_velocity, _particle_position

    for axi, ax in enumerate("xyz"):
        v, p = _get_coord_funcs(axi, ptype)
        registry.add_field(
            (ptype, f"particle_velocity_{ax}"),
            sampling_type="particle",
            function=v,
            units="code_velocity",
        )
        registry.add_field(
            (ptype, f"particle_position_{ax}"),
            sampling_type="particle",
            function=p,
            units="code_length",
        )


def particle_vector_functions(ptype, coord_names, vel_names, registry):

    unit_system = registry.ds.unit_system

    # This will column_stack a set of scalars to create vector fields.

    def _get_vec_func(_ptype, names):
        def particle_vectors(field, data):
            v = [data[_ptype, name].in_units(field.units) for name in names]
            return data.ds.arr(np.column_stack(v), v[0].units)

        return particle_vectors

    registry.add_field(
        (ptype, "particle_position"),
        sampling_type="particle",
        function=_get_vec_func(ptype, coord_names),
        units="code_length",
    )

    registry.add_field(
        (ptype, "particle_velocity"),
        sampling_type="particle",
        function=_get_vec_func(ptype, vel_names),
        units=unit_system["velocity"],
    )


def get_angular_momentum_components(ptype, data, spos, svel):
    if data.has_field_parameter("normal"):
        normal = data.get_field_parameter("normal")
    else:
        normal = data.ds.arr(
            [0.0, 0.0, 1.0], "code_length"
        )  # default to simulation axis
    pos = data.ds.arr([data[ptype, spos % ax] for ax in "xyz"]).T
    vel = data.ds.arr([data[ptype, f"relative_{svel % ax}"] for ax in "xyz"]).T
    return pos, vel, normal


def standard_particle_fields(
    registry, ptype, spos="particle_position_%s", svel="particle_velocity_%s"
):
    unit_system = registry.ds.unit_system

    def _particle_velocity_magnitude(field, data):
        """M{|v|}"""
        return np.sqrt(
            data[ptype, f"relative_{svel % 'x'}"] ** 2
            + data[ptype, f"relative_{svel % 'y'}"] ** 2
            + data[ptype, f"relative_{svel % 'z'}"] ** 2
        )

    registry.add_field(
        (ptype, "particle_velocity_magnitude"),
        sampling_type="particle",
        function=_particle_velocity_magnitude,
        take_log=False,
        units=unit_system["velocity"],
    )

    def _particle_specific_angular_momentum(field, data):
        """Calculate the angular of a particle velocity.

        Returns a vector for each particle.
        """
        center = data.get_field_parameter("center")
        pos, vel, normal = get_angular_momentum_components(ptype, data, spos, svel)
        L, r_vec, v_vec = modify_reference_frame(center, normal, P=pos, V=vel)
        # adding in the unit registry allows us to have a reference to the
        # dataset and thus we will always get the correct units after applying
        # the cross product.
        return ucross(r_vec, v_vec, registry=data.ds.unit_registry)

    registry.add_field(
        (ptype, "particle_specific_angular_momentum"),
        sampling_type="particle",
        function=_particle_specific_angular_momentum,
        units=unit_system["specific_angular_momentum"],
        validators=[ValidateParameter("center")],
    )

    def _get_spec_ang_mom_comp(axi, ax, _ptype):
        def _particle_specific_angular_momentum_component(field, data):
            return data[_ptype, "particle_specific_angular_momentum"][:, axi]

        def _particle_angular_momentum_component(field, data):
            return (
                data[_ptype, "particle_mass"]
                * data[ptype, f"particle_specific_angular_momentum_{ax}"]
            )

        return (
            _particle_specific_angular_momentum_component,
            _particle_angular_momentum_component,
        )

    for axi, ax in enumerate("xyz"):
        f, v = _get_spec_ang_mom_comp(axi, ax, ptype)
        registry.add_field(
            (ptype, f"particle_specific_angular_momentum_{ax}"),
            sampling_type="particle",
            function=f,
            units=unit_system["specific_angular_momentum"],
            validators=[ValidateParameter("center")],
        )
        registry.add_field(
            (ptype, f"particle_angular_momentum_{ax}"),
            sampling_type="particle",
            function=v,
            units=unit_system["angular_momentum"],
            validators=[ValidateParameter("center")],
        )

    def _particle_angular_momentum(field, data):
        am = (
            data[ptype, "particle_mass"]
            * data[ptype, "particle_specific_angular_momentum"].T
        )
        return am.T

    registry.add_field(
        (ptype, "particle_angular_momentum"),
        sampling_type="particle",
        function=_particle_angular_momentum,
        units=unit_system["angular_momentum"],
        validators=[ValidateParameter("center")],
    )

    create_magnitude_field(
        registry,
        "particle_angular_momentum",
        unit_system["angular_momentum"],
        sampling_type="particle",
        ftype=ptype,
    )

    def _particle_radius(field, data):
        """The spherical radius component of the particle positions

        Relative to the coordinate system defined by the *normal* vector,
        and *center* field parameters.
        """
        return get_radius(data, "particle_position_", field.name[0])

    registry.add_field(
        (ptype, "particle_radius"),
        sampling_type="particle",
        function=_particle_radius,
        units=unit_system["length"],
        validators=[ValidateParameter("center")],
    )

    def _relative_particle_position(field, data):
        """The cartesian particle positions in a rotated reference frame

        Relative to the coordinate system defined by *center* field parameter.

        Note that the orientation of the x and y axes are arbitrary.
        """
        field_names = [(ptype, f"particle_position_{ax}") for ax in "xyz"]
        return obtain_position_vector(data, field_names=field_names).T

    registry.add_field(
        (ptype, "relative_particle_position"),
        sampling_type="particle",
        function=_relative_particle_position,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    def _relative_particle_velocity(field, data):
        """The vector particle velocities in an arbitrary coordinate system

        Relative to the coordinate system defined by the *bulk_velocity*
        vector field parameter.

        Note that the orientation of the x and y axes are arbitrary.
        """
        field_names = [(ptype, f"particle_velocity_{ax}") for ax in "xyz"]
        return obtain_relative_velocity_vector(data, field_names=field_names).T

    registry.add_field(
        (ptype, "relative_particle_velocity"),
        sampling_type="particle",
        function=_relative_particle_velocity,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    def _get_coord_funcs_relative(axi, _ptype):
        def _particle_pos_rel(field, data):
            return data[_ptype, "relative_particle_position"][:, axi]

        def _particle_vel_rel(field, data):
            return data[_ptype, "relative_particle_velocity"][:, axi]

        return _particle_vel_rel, _particle_pos_rel

    for axi, ax in enumerate("xyz"):
        v, p = _get_coord_funcs_relative(axi, ptype)
        registry.add_field(
            (ptype, f"particle_velocity_relative_{ax}"),
            sampling_type="particle",
            function=v,
            units="code_velocity",
        )
        registry.add_field(
            (ptype, f"particle_position_relative_{ax}"),
            sampling_type="particle",
            function=p,
            units="code_length",
        )
        registry.add_field(
            (ptype, f"relative_particle_velocity_{ax}"),
            sampling_type="particle",
            function=v,
            units="code_velocity",
        )
        registry.add_field(
            (ptype, f"relative_particle_position_{ax}"),
            sampling_type="particle",
            function=p,
            units="code_length",
        )

    # this is just particle radius but we add it with an alias for the sake of
    # consistent naming
    registry.add_field(
        (ptype, "particle_position_spherical_radius"),
        sampling_type="particle",
        function=_particle_radius,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_spherical_position_radius"),
        (ptype, "particle_position_spherical_radius"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_position_spherical_theta(field, data):
        """The spherical theta coordinate of the particle positions.

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        return data.ds.arr(get_sph_theta(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_spherical_theta"),
        sampling_type="particle",
        function=_particle_position_spherical_theta,
        units="",
        validators=[ValidateParameter("center"), ValidateParameter("normal")],
    )

    registry.alias(
        (ptype, "particle_spherical_position_theta"),
        (ptype, "particle_position_spherical_theta"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_position_spherical_phi(field, data):
        """The spherical phi component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        return data.ds.arr(get_sph_phi(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_spherical_phi"),
        sampling_type="particle",
        function=_particle_position_spherical_phi,
        units="",
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_spherical_position_phi"),
        (ptype, "particle_position_spherical_phi"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_velocity_spherical_radius(field, data):
        """The spherical radius component of the particle velocities in an
         arbitrary coordinate system

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        vel = data[(ptype, "relative_particle_velocity")].T
        theta = get_sph_theta(pos, normal)
        phi = get_sph_phi(pos, normal)
        sphr = get_sph_r_component(vel, theta, phi, normal)
        return sphr

    registry.add_field(
        (ptype, "particle_velocity_spherical_radius"),
        sampling_type="particle",
        function=_particle_velocity_spherical_radius,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_spherical_velocity_radius"),
        (ptype, "particle_velocity_spherical_radius"),
        deprecate=("4.0.0", "4.1.0"),
    )

    registry.alias(
        (ptype, "particle_radial_velocity"),
        (ptype, "particle_velocity_spherical_radius"),
    )

    def _particle_velocity_spherical_theta(field, data):
        """The spherical theta component of the particle velocities in an
         arbitrary coordinate system

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        vel = data[(ptype, "relative_particle_velocity")].T
        theta = get_sph_theta(pos, normal)
        phi = get_sph_phi(pos, normal)
        spht = get_sph_theta_component(vel, theta, phi, normal)
        return spht

    registry.add_field(
        (ptype, "particle_velocity_spherical_theta"),
        sampling_type="particle",
        function=_particle_velocity_spherical_theta,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_spherical_velocity_theta"),
        (ptype, "particle_velocity_spherical_theta"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_velocity_spherical_phi(field, data):
        """The spherical phi component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        vel = data[(ptype, "relative_particle_velocity")].T
        phi = get_sph_phi(pos, normal)
        sphp = get_sph_phi_component(vel, phi, normal)
        return sphp

    registry.add_field(
        (ptype, "particle_velocity_spherical_phi"),
        sampling_type="particle",
        function=_particle_velocity_spherical_phi,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_spherical_velocity_phi"),
        (ptype, "particle_velocity_spherical_phi"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_position_cylindrical_radius(field, data):
        """The cylindrical radius component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        pos.convert_to_units("code_length")
        return data.ds.arr(get_cyl_r(pos, normal), "code_length")

    registry.add_field(
        (ptype, "particle_position_cylindrical_radius"),
        sampling_type="particle",
        function=_particle_position_cylindrical_radius,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    def _particle_position_cylindrical_theta(field, data):
        """The cylindrical theta component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        return data.ds.arr(get_cyl_theta(pos, normal), "")

    registry.add_field(
        (ptype, "particle_position_cylindrical_theta"),
        sampling_type="particle",
        function=_particle_position_cylindrical_theta,
        units="",
        validators=[ValidateParameter("center"), ValidateParameter("normal")],
    )

    def _particle_position_cylindrical_z(field, data):
        """The cylindrical z component of the particle positions

        Relative to the coordinate system defined by the *normal* vector
        and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        pos.convert_to_units("code_length")
        return data.ds.arr(get_cyl_z(pos, normal), "code_length")

    registry.add_field(
        (ptype, "particle_position_cylindrical_z"),
        sampling_type="particle",
        function=_particle_position_cylindrical_z,
        units=unit_system["length"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    def _particle_velocity_cylindrical_radius(field, data):
        """The cylindrical radius component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        vel = data[(ptype, "relative_particle_velocity")].T
        theta = get_cyl_theta(pos, normal)
        cylr = get_cyl_r_component(vel, theta, normal)
        return cylr

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_radius"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_radius,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    def _particle_velocity_cylindrical_theta(field, data):
        """The cylindrical theta component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        pos = data[(ptype, "relative_particle_position")].T
        vel = data[(ptype, "relative_particle_velocity")].T
        theta = get_cyl_theta(pos, normal)
        cylt = get_cyl_theta_component(vel, theta, normal)
        return cylt

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_theta"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_theta,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_cylindrical_velocity_theta"),
        (ptype, "particle_velocity_cylindrical_theta"),
        deprecate=("4.0.0", "4.1.0"),
    )

    def _particle_velocity_cylindrical_z(field, data):
        """The cylindrical z component of the particle velocities

        Relative to the coordinate system defined by the *normal* vector,
        *bulk_velocity* vector and *center* field parameters.
        """
        normal = data.get_field_parameter("normal")
        vel = data[(ptype, "relative_particle_velocity")].T
        cylz = get_cyl_z_component(vel, normal)
        return cylz

    registry.add_field(
        (ptype, "particle_velocity_cylindrical_z"),
        sampling_type="particle",
        function=_particle_velocity_cylindrical_z,
        units=unit_system["velocity"],
        validators=[ValidateParameter("normal"), ValidateParameter("center")],
    )

    registry.alias(
        (ptype, "particle_cylindrical_velocity_z"),
        (ptype, "particle_velocity_cylindrical_z"),
        deprecate=("4.0.0", "4.1.0"),
    )


def add_particle_average(registry, ptype, field_name, weight=None, density=True):
    if weight is None:
        weight = (ptype, "particle_mass")
    field_units = registry[ptype, field_name].units

    def _pfunc_avg(field, data):
        pos = data[ptype, "particle_position"]
        f = data[ptype, field_name]
        wf = data[ptype, weight]
        f *= wf
        v = data.deposit(pos, [f], method="sum")
        w = data.deposit(pos, [wf], method="sum")
        v /= w
        if density:
            v /= data["index", "cell_volume"]
        v[np.isnan(v)] = 0.0
        return v

    fn = ("deposit", f"{ptype}_avg_{field_name}")
    registry.add_field(
        fn,
        sampling_type="cell",
        function=_pfunc_avg,
        validators=[ValidateSpatial(0)],
        units=field_units,
    )
    return fn


def add_volume_weighted_smoothed_field(
    ptype,
    coord_name,
    mass_name,
    smoothing_length_name,
    density_name,
    smoothed_field,
    registry,
    nneighbors=64,
    kernel_name="cubic",
):
    from yt._maintenance.deprecation import issue_deprecation_warning

    issue_deprecation_warning(
        "This function is deprecated. "
        "Since yt-4.0, it's no longer necessary to add a field specifically for "
        "smoothing, because the global octree is removed. The old behavior of "
        "interpolating onto a grid structure can be recovered through data objects "
        "like ds.arbitrary_grid, ds.covering_grid, and most closely ds.octree. The "
        "visualization machinery now treats SPH fields properly by smoothing onto "
        "pixel locations. See this page to learn more: "
        "https://yt-project.org/doc/yt4differences.html",
        since="4.0.0",
        removal="4.1.0",
    )


def add_nearest_neighbor_field(ptype, coord_name, registry, nneighbors=64):
    field_name = (ptype, f"nearest_neighbor_distance_{nneighbors}")

    def _nth_neighbor(field, data):
        pos = data[ptype, coord_name]
        pos.convert_to_units("code_length")
        distances = 0.0 * pos[:, 0]
        data.particle_operation(
            pos, [distances], method="nth_neighbor", nneighbors=nneighbors
        )
        # Now some quick unit conversions.
        return distances

    registry.add_field(
        field_name,
        sampling_type="particle",
        function=_nth_neighbor,
        validators=[ValidateSpatial(0)],
        units="code_length",
    )
    return [field_name]


def add_nearest_neighbor_value_field(ptype, coord_name, sampled_field, registry):
    """
    This adds a nearest-neighbor field, where values on the mesh are assigned
    based on the nearest particle value found.  This is useful, for instance,
    with voronoi-tesselations.
    """
    field_name = ("deposit", f"{ptype}_nearest_{sampled_field}")
    field_units = registry[ptype, sampled_field].units
    unit_system = registry.ds.unit_system

    def _nearest_value(field, data):
        pos = data[ptype, coord_name]
        pos = pos.convert_to_units("code_length")
        value = data[ptype, sampled_field].in_base(unit_system.name)
        rv = data.smooth(
            pos, [value], method="nearest", create_octree=True, nneighbors=1
        )
        rv = data.apply_units(rv, field_units)
        return rv

    registry.add_field(
        field_name,
        sampling_type="cell",
        function=_nearest_value,
        validators=[ValidateSpatial(0)],
        units=field_units,
    )
    return [field_name]


def add_union_field(registry, ptype, field_name, units):
    """
    Create a field that is the concatenation of multiple particle types.
    This allows us to create fields for particle unions using alias names.
    """

    def _cat_field(field, data):
        return uconcatenate(
            [data[dep_type, field_name] for dep_type in data.ds.particle_types_raw]
        )

    registry.add_field(
        (ptype, field_name), sampling_type="particle", function=_cat_field, units=units
    )
